from pusion.core.combiner import TrainableCombiner
from pusion.util.transformer import *
from pusion.util.constants import *


class DecisionTemplatesCombiner(TrainableCombiner):
    """
    The :class:`DecisionTemplatesCombiner` (DT) is adopted from the decision fusion method originally proposed by
    Kuncheva :footcite:`kuncheva2014combining`. A decision template is the average matrix of all decision profiles,
    which correspond to samples of one specific class. A decision profile contains classification outputs from all
    classifiers for a sample in a row-wise fashion. The decision fusion is performed based on distance calculations
    between decision templates and the decision profile generated from the ensemble outputs.

    .. footbibliography::
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'DT'

    def __init__(self):
        super().__init__()
        # all possible classification occurrences (class assignments) in training data
        self.distinct_class_assignments = None
        # decision templates according to kuncheva per distinct class assignment (aligned with
        # distinct_class_assignments)
        self.decision_templates = None

    def train(self, decision_tensor, true_assignments):
        """
        Train the Decision Templates Combiner model by precalculating decision templates from given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported. This procedure involves
        calculating means of decision profiles (decision templates) for each true class.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        if np.shape(decision_tensor)[1] != np.shape(true_assignments)[0]:
            raise TypeError("True assignment vector dimension does not match the number of samples.")

        # represent outputs of multiple classifiers as a DP for each sample
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        self.distinct_class_assignments = np.unique(true_assignments, axis=0)

        self.decision_templates = np.zeros((len(self.distinct_class_assignments),
                                            np.shape(decision_profiles[0])[0],
                                            np.shape(decision_profiles[0])[1]))

        for i in range(len(self.distinct_class_assignments)):
            # calculate the mean decision profile (decision template) for each class assignment.
            label = self.distinct_class_assignments[i]
            label_indices = np.where(np.all(label == true_assignments, axis=1))[0]
            self.decision_templates[i] = np.average(decision_profiles[label_indices], axis=0)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by using the Decision Templates method.
        Both continuous and crisp classification outputs are supported. Combining requires a trained
        :class:`DecisionTemplatesCombiner`.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by the minimum distance between decision profiles of ``decision_tensor`` and
                precalculated decision templates. Axis 0 represents samples and axis 1 the class assignments which
                are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        fused_decisions = np.zeros_like(decision_tensor[0])

        # Compute the euclidean distance between each DP (for a class assignment) and trained DT.
        # The class assignment associated with the DT with minimal distance to the DP
        # is considered as the fused decision.
        for i in range(len(decision_profiles)):
            dp = decision_profiles[i]
            dist = np.empty(len(self.decision_templates))
            for j in range(len(self.decision_templates)):
                dt = self.decision_templates[j]
                dist[j] = np.average((dp - dt)**2)
            min_dist_label = self.distinct_class_assignments[dist.argmin()]
            fused_decisions[i] = min_dist_label
        return fused_decisions

    def get_decision_templates(self):
        return self.decision_templates

    def get_distinct_labels(self):
        return self.distinct_class_assignments


class CRDecisionTemplatesCombiner(DecisionTemplatesCombiner):
    """
    The :class:`CRDecisionTemplatesCombiner` is a modification of :class:`DecisionTemplatesCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as a constant, respectively. To use methods :meth:`train` and
    :meth:`combine` a coverage needs to be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
    ]

    def __init__(self):
        super().__init__()

    def train(self, decision_outputs, true_assignments):
        """
        Train the Decision Templates Combiner model by precalculating decision templates from given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported. This procedure involves
        calculating means of decision profiles (decision templates) for each true class.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains either crisp or continuous
                decision outputs per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        super().train(t_decision_outputs, true_assignments)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by using the Decision Templates method.
        Both continuous and crisp classification outputs are supported. Combining requires a trained
        :class:`CRDecisionTemplatesCombiner`.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of crisp or continuous class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.negative(np.ones((n_classifiers, n_decisions, n_classes)))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



from pusion.core.combiner import EvidenceBasedCombiner, TrainableCombiner
from pusion.util.generator import *
from pusion.evaluation.evaluation_metrics import accuracy
from pusion.util.constants import *


class WeightedVotingCombiner(EvidenceBasedCombiner, TrainableCombiner):
    """
    The :class:`WeightedVotingCombiner` (WV) is a weighted voting schema adopted from Kuncheva (eq. 4.43)
    :footcite:`kuncheva2014combining`. Classifiers with better performance (i.e. accuracy) are given more
    weight contributing to final decisions. Nevertheless, if classifiers of high performance disagree on a sample,
    low performance classifiers may contribute to the final decision.

    .. footbibliography::
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'WV'

    def __init__(self):
        super().__init__()
        self.accuracy = None

    def set_evidence(self, evidence):
        """
        Set the evidence given by confusion matrices calculated according to Kuncheva :footcite:`kuncheva2014combining`
        for each ensemble classifier.

        .. footbibliography::

        :param evidence: `numpy.array` of shape `(n_classifiers, n_classes, n_classes)`.
                Confusion matrices for each of `n` classifiers.
        """
        self.accuracy = confusion_matrices_to_accuracy_vector(evidence)

    def train(self, decision_tensor, true_assignments):
        """
        Train the Weighted Voting combiner model by precalculating confusion matrices from given decision outputs and
        true class assignments. Continuous decision outputs are converted into crisp multiclass assignments using
        the MAX rule.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        cms = generate_multiclass_confusion_matrices(decision_tensor, true_assignments)
        self.accuracy = confusion_matrices_to_accuracy_vector(cms)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by the weighted voting schema.
        Classifiers with better performance (i.e. accuracy) are given more authority over final decisions.
        Combining requires a trained :class:`WeightedVotingCombiner` or evidence set with ``set_evidence``.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which represents fused
                decisions obtained by the maximum weighted class support. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        if self.accuracy is None:
            raise TypeError("Accuracy is not set for this model as an evidence.")

        if np.shape(decision_tensor)[0] != np.shape(self.accuracy)[0]:
            raise TypeError("Accuracy vector dimension does not match the number of classifiers in the input tensor.")
        # convert decision_tensor to decision profiles for better handling
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        # average all decision profiles, i-th row contains decisions for i-th sample
        decision_matrix = np.array([np.average(dp, axis=0, weights=self.accuracy) for dp in decision_profiles])
        fused_decisions = np.zeros_like(decision_matrix)
        # find the maximum class support according to Kuncheva eq. (4.43)
        fused_decisions[np.arange(len(fused_decisions)), decision_matrix.argmax(axis=1)] = 1
        return fused_decisions


class CRWeightedVotingCombiner(WeightedVotingCombiner):
    """
    The :class:`CRWeightedVotingCombiner` is a modification of :class:`WeightedVotingCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed to a unified
    tensor representation supporting undefined class assignments. The mean is calculated only for assignments which
    are defined. To call methods :meth:`train` and :meth:`combine`, a coverage needs to be set first
    by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
    ]

    def __init__(self):
        super().__init__()
        self.accuracy = None

    def train(self, decision_outputs, true_assignments):
        """
        Train the Weighted Voting combiner model by precalculating confusion matrices from given decision outputs and
        true class assignments. Continuous decision outputs are converted into crisp multiclass assignments using
        the MAX rule.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains crisp decision outputs
                per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which is considered true for each sample during
                the training procedure.
        """
        self.accuracy = np.zeros(len(decision_outputs))
        for i in range(len(decision_outputs)):
            y_true = true_assignments[:, self.coverage[i]]
            y_pred = decision_outputs[i]
            self.accuracy[i] = accuracy(y_true, y_pred)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by the weighted voting schema.
        Classifiers with better performance (i.e. accuracy) are given more authority over final decisions.
        Combining requires a trained :class:`WeightedVotingCombiner` or evidence set with ``set_evidence``.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains crisp decision outputs
                per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which are obtained by the best representative class
                for a certain classifier's behaviour per sample. Axis 0 represents samples and axis 1 all the class
                labels which are provided by the coverage.
        """
        if self.accuracy is None:
            raise TypeError("Accuracy is not set for this model as an evidence.")
        if len(decision_outputs) != len(self.accuracy):
            raise TypeError("Accuracy vector dimension does not match the number of classifiers in the input tensor.")

        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)

        # convert decision_tensor to decision profiles for better handling
        decision_profiles = decision_tensor_to_decision_profiles(t_decision_outputs)
        # use a masked array due to classifications which do not cover all classes
        masked_decision_profiles = np.ma.masked_array(decision_profiles, np.isnan(decision_profiles))
        # average all decision templates, i-th row contains decisions of i-th sample
        adp = np.array([np.ma.average(mdp, axis=0, weights=self.accuracy) for mdp in masked_decision_profiles])
        # adp = np.array([ad.filled(np.nan) for ad in adp])  # coverage test should avoid nan columns

        fused_decisions = np.zeros_like(adp)
        # find the maximum class support according to Kuncheva eq. (4.43)
        fused_decisions[np.arange(len(fused_decisions)), adp.argmax(1)] = 1
        return fused_decisions

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.full((n_classifiers, n_decisions, n_classes), np.nan)
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



__all__ = ["behaviour_knowledge_space_combiner",
           "borda_count_combiner",
           "combiner",
           "complementary_output_combiner",
           "cosine_similarity_combiner",
           "decision_templates_combiner",
           "dempster_shafer_combiner",
           "k_nearest_neighbors_combiner",
           "macro_majority_vote_combiner",
           "micro_majority_vote_combiner",
           "naive_bayes_combiner",
           "neural_network_combiner",
           "simple_average_combiner",
           "weighted_voting_combiner"]


import numpy as np


from pusion.core.combiner import UtilityBasedCombiner
from pusion.util.constants import *


class ComplementaryOutputCombiner(UtilityBasedCombiner):
    """
    The :class:`ComplementaryOutputCombiner` combines fully complementary decision outputs by concatenating individual
    decisions across classes for each sample.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY),
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY)
    ]

    SHORT_NAME = 'COB'

    def __init__(self):
        UtilityBasedCombiner.__init__(self)

    def combine(self, decision_outputs):
        """
        Combine fully complementary decision outputs by concatenating individual decisions according to the coverage
        of all classifiers. Due to the nature of complementary class coverage, no fusion between redundant class
        assignments is required.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by the highest cumulative cosine-similarity. Axis 0 represents samples and axis 1 the
                class labels which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        n_classes = np.sum([len(ca) for ca in self.coverage])
        fused_decisions = np.zeros_like((len(decision_outputs[0]), n_classes))

        for i, classifier_coverage in enumerate(self.coverage):
            for ci in classifier_coverage:
                fused_decisions[:, ci] = decision_outputs[i, :, ci]
        return fused_decisions



from pusion.auto.auto_combiner import AutoCombiner
from pusion.auto.generic_combiner import GenericCombiner
from pusion.control.decision_processor import DecisionProcessor
from pusion.core import *
from pusion.core.behaviour_knowledge_space_combiner import BehaviourKnowledgeSpaceCombiner
from pusion.core.borda_count_combiner import BordaCountCombiner
from pusion.core.cosine_similarity_combiner import CosineSimilarityCombiner
from pusion.core.decision_templates_combiner import DecisionTemplatesCombiner
from pusion.core.k_nearest_neighbors_combiner import KNNCombiner
from pusion.core.dempster_shafer_combiner import DempsterShaferCombiner
from pusion.core.macro_majority_vote_combiner import MacroMajorityVoteCombiner
from pusion.core.maximum_likelihood_combiner import MaximumLikelihoodCombiner
from pusion.core.micro_majority_vote_combiner import MicroMajorityVoteCombiner
from pusion.core.naive_bayes_combiner import NaiveBayesCombiner
from pusion.core.neural_network_combiner import NeuralNetworkCombiner
from pusion.core.simple_average_combiner import SimpleAverageCombiner
from pusion.core.weighted_voting_combiner import WeightedVotingCombiner
from pusion.evaluation.evaluation import Evaluation
from pusion.evaluation.evaluation_metrics import *
from pusion.model.configuration import Configuration
from pusion.util import *
from pusion.util.constants import *
from pusion.util.exceptions import *
from pusion.util.generator import *

# Maintain this static attributes according to implemented methods for ease of framework usage,
# e.g., `pusion.Method.AutoCombiner`.
class Method:
    BEHAVIOUR_KNOWLEDGE_SPACE = BehaviourKnowledgeSpaceCombiner
    BORDA_COUNT = BordaCountCombiner
    COSINE_SIMILARITY = CosineSimilarityCombiner
    DECISION_TEMPLATES = DecisionTemplatesCombiner
    DEMPSTER_SHAFER = DempsterShaferCombiner
    KNN = KNNCombiner
    MACRO_MAJORITY_VOTE = MacroMajorityVoteCombiner
    MAXIMUM_LIKELIHOOD = MaximumLikelihoodCombiner
    MICRO_MAJORITY_VOTE = MicroMajorityVoteCombiner
    NAIVE_BAYES = NaiveBayesCombiner
    NEURAL_NETWORK = NeuralNetworkCombiner
    SIMPLE_AVERAGE = SimpleAverageCombiner
    WEIGHTED_VOTING = WeightedVotingCombiner
    AUTO = AutoCombiner
    GENERIC = GenericCombiner


class PerformanceMetric:
    MICRO_PRECISION = micro_precision
    MICRO_RECALL = micro_recall
    MICRO_F1_SCORE = micro_f1
    MICRO_F2_SCORE = micro_f2
    MICRO_JACCARD_SCORE = micro_jaccard
    MACRO_PRECISION = macro_precision
    MACRO_RECALL = macro_recall
    MACRO_F1_SCORE = macro_f1
    MACRO_F2_SCORE = macro_f2
    WEIGHTED_F1_SCORE = weighted_f1
    MACRO_JACCARD_SCORE = macro_jaccard
    WEIGTED_JACCARD_SCORE = weighted_jaccard
    HAMMING_LOSS = hamming
    ACCURACY = accuracy
    MEAN_MULTILABEL_ACCURACY = mean_multilabel_confusion_matrix
    MEAN_CONFIDENCE = mean_confidence
    BALANCED_MULTICLASS_ACCURACY_SCORE = balanced_multiclass_accuracy
    ERROR_RATE = error_rate
    MULTILABEL_BRIER_SCORE_MICRO = multi_label_brier_score_micro
    MULTILABEL_BRIER_SCORE = multi_label_brier_score
    MULTICLASS_BRIER_SCORE = multiclass_brier_score
    FALSE_ALARM_RATE = far
    MULTICLASS_FDR = multiclass_fdr
    MULTILABEL_SUBSET_FDR = multilabel_subset_fdr
    MULTILABEL_MINOR_FDR = multilabel_minor_fdr
    MULTICLASS_WEIGHTED_PRECISION = multiclass_weighted_precision
    MULTILABEL_WEIGHTED_PRECISION = multi_label_weighted_precision
    MULTICLASS_CLASS_WISE_PRECISION = multiclass_class_wise_precision
    MULTILABEL_CLASS_WISE_PRECISION = multi_label_class_wise_precision
    MULTICLASS_RECALL = multiclass_recall
    MULTILABEL_RECALL = multi_label_recall
    MULTICLASS_CLASS_WISE_RECALL = multiclass_class_wise_recall
    MULTILABEL_CLASS_WISE_RECALL = multi_label_class_wise_recall
    MULTICLASS_WEIGHTED_SCIKIT_AUC_ROC_SCORE = multiclass_weighted_scikit_auc_roc_score
    MULTILABEL_WEIGHTED_PYTORCH_AUC_ROC_SCORE = multi_label_weighted_pytorch_auc_roc_score
    MULTILABEL_PYTORCH_AUC_ROC_SCORE = multi_label_pytorch_auc_roc_score
    MULTICLASS_CLASS_WISE_AVG_PRECISION = multiclass_class_wise_avg_precision
    MULTICLASS_WEIGHTED_AVG_PRECISION = multiclass_weighted_avg_precision
    MULTICLASS_AUC_PRECISION_RECALL_CURVE = multiclass_auc_precision_recall_curve
    MULTICLASS_WEIGHTED_PYTORCH_AUC_ROC = multiclass_weighted_pytorch_auc_roc
    MULTICLASS_PYTORCH_AUC_ROC = multiclass_pytorch_auc_roc
    MULTILABEL_RANKING_AVG_PRECISION_SCORE = multi_label_ranking_avg_precision_score
    MULTILABEL_RANKING_LOSS = multi_label_ranking_loss
    MULTILABEL_NORMALIZED_DISCOUNTED_CUMULATIVE_GAIN = multi_label_normalized_discounted_cumulative_gain
    MULTICLASS_TOP_1_ACCURACY = multiclass_top_1_accuracy
    MULTICLASS_TOP_3_ACCURACY = multiclass_top_3_accuracy
    MULTICLASS_TOP_5_ACCURACY = multiclass_top_5_accuracy
    MULTICLASS_LOG_LOSS = multiclass_log_loss
    MULTILABEL_LOG_LOSS = multi_label_log_loss



class DiversityMetric:
    PAIRWISE_COHENS_KAPPA = pairwise_cohens_kappa
    PAIRWISE_CORRELATION = pairwise_correlation
    PAIRWISE_Q_STATISTIC = pairwise_q_statistic
    PAIRWISE_KAPPA_STATISTIC = pairwise_kappa_statistic
    PAIRWISE_DISAGREEMENT = pairwise_disagreement
    PAIRWISE_DOUBLE_FAULT = pairwise_double_fault
    PAIRWISE_EUCLIDEAN_DISTANCE = pairwise_euclidean_distance



from pusion.util.constants import *
from pusion.util.transformer import *


def determine_problem(decision_outputs):
    """
    Determine the classification problem based on the decision outputs.

    :param decision_outputs: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
            `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
            due to the coverage.
    :return: `string` constant `'MULTI_CLASS'` or `'MULTI_LABEL'`. See `pusion.util.constants.Problem`.
    """
    if isinstance(decision_outputs, np.ndarray):
        return __determine_tensor_problem(decision_outputs)

    problem_list = []
    for dt in decision_outputs:
        problem_list.append(__determine_tensor_problem(np.array(dt)))
    if np.all(np.char.strip(problem_list) == Problem.MULTI_CLASS):
        return Problem.MULTI_CLASS
    return Problem.MULTI_LABEL


def __determine_tensor_problem(decision_tensor):
    """
    Helper method for ``determine_problem`` to determine the classification problem on `numpy.array` tensors.
    """
    # sum up all decisions along classes tolerating the inaccuracy of floating point numbers
    decision_sum = np.around(np.sum(decision_tensor, axis=decision_tensor.ndim - 1), 5)
    assignment_type = determine_assignment_type(decision_tensor)
    if assignment_type == AssignmentType.CRISP:
        if np.all(decision_sum == 1):
            return Problem.MULTI_CLASS
        if np.all(decision_sum >= 0):  # >= 1, if zero outputs should not be tolerated by framework requirements.
            return Problem.MULTI_LABEL

    if assignment_type == AssignmentType.CONTINUOUS:
        if np.all(decision_sum == 1):
            return Problem.MULTI_CLASS
        else:
            return Problem.MULTI_LABEL
    raise TypeError("The problem could not be determined for the given decision tensor.")


def determine_assignment_type(decision_outputs):
    """
    Determine the assignment type based on the decision outputs.

    :param decision_outputs: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
            `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
            due to the coverage.
    :return: `string` constant `'CRISP'` or `'CONTINUOUS'`. See `pusion.util.constants.AssignmentType`.
    """
    if isinstance(decision_outputs, np.ndarray):
        return __determine_tensor_assignment_type(decision_outputs)

    assignment_type_list = []
    for dt in decision_outputs:
        assignment_type_list.append(__determine_tensor_assignment_type(np.array(dt)))
    if np.all(np.char.strip(assignment_type_list) == AssignmentType.CRISP):
        return AssignmentType.CRISP
    return AssignmentType.CONTINUOUS


def __determine_tensor_assignment_type(decision_tensor):
    """
    Helper method for ``determine_assignment_type`` to determine the assignment type on `numpy.array` tensors.
    """
    if np.any(decision_tensor < 0) or np.any(decision_tensor > 1) or np.all(decision_tensor == 0):
        raise TypeError("The assignment type could not be determined for the given decision tensor.")

    if np.all(np.logical_or(decision_tensor == 0, decision_tensor == 1)):
        return AssignmentType.CRISP

    if np.all(np.logical_and(decision_tensor >= 0, decision_tensor <= 1)):
        return AssignmentType.CONTINUOUS


def determine_coverage_type(coverage):
    """
    Determine the coverage type.

    :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a classifier,
            which is identified by the positional index of the respective list.
    :return: `string` constant `'REDUNDANT'`, `'COMPLEMENTARY'`  or `'COMPLEMENTARY_REDUNDANT'`.
            See `pusion.util.constants.CoverageType`.
    """

    # determine coverage sets
    omega = [set(classes) for classes in coverage]
    # check for redundancy
    equality = [omega[i] == omega[j] for i in range(len(omega)) for j in range(i)]
    if np.all(equality):
        return CoverageType.REDUNDANT
    # check for complementary
    disjunction = [omega[i].intersection(omega[j]) == set() for i in range(len(omega)) for j in range(i)]
    if np.all(disjunction):
        return CoverageType.COMPLEMENTARY
    # complementary-redundant, since neither redundancy nor complementary rules apply
    return CoverageType.COMPLEMENTARY_REDUNDANT


def determine_pac(decision_outputs, coverage=None):
    """
    Determine the PAC-tuple (problem, assignment type and coverage type) based on the given decision outputs and
    coverage.

    :param decision_outputs: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
            `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
            due to the coverage.
    :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a classifier,
            which is identified by the positional index of the respective list.
    :return: `tuple` of string constants representing the PAC. See `pusion.util.constants.Problem`,
            `pusion.util.constants.AssignmentType` and `pusion.util.constants.CoverageType`.
    """
    if not coverage:
        if not isinstance(decision_outputs, np.ndarray):
            raise TypeError('A non-numpy array indicating a possible complementary or complementary-redundant coverage '
                            'is given without coverage specification.')
        return determine_problem(decision_outputs), determine_assignment_type(decision_outputs), CoverageType.REDUNDANT
    else:
        return determine_problem(decision_outputs), \
               determine_assignment_type(decision_outputs), \
               determine_coverage_type(coverage)



from abc import abstractmethod

from pusion.util.exceptions import ValidationError


class Combiner:
    """
    Combiner's root class. This class as well as the subclasses in this module set the structure for each combiner
    provided by the framework. It also accommodates methods and attributes which are essential for combiner's
    registration and obtainment.

    Each combiner to be registered in the framework, needs to base at least one of the following classes:

    - :class:`UtilityBasedCombiner`
    - :class:`TrainableCombiner`
    - :class:`EvidenceBasedCombiner`.

    Furthermore, it needs a `list` :attr:`_SUPPORTED_PAC` of supported PAC tuples set as a class attribute.
    A PAC tuple is a `tuple` of string constants (classification problem, assignment type and coverage type). See:

    - `pusion.util.constants.Problem`
    - `pusion.util.constants.AssignmentType`
    - `pusion.util.constants.CoverageType`

    respectively. Example of a new combiner:

    .. code:: python

        class NewCombiner(TrainableCombiner):

        _SUPPORTED_PAC = [
            (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
            (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT)
        ]

        def train(self, decision_outputs, true_assignments):
            pass

        def combine(self, decision_outputs):
            pass

    .. warning::

        Note that a new combiner also needs to be inserted into the :class:`pusion.Method` class within
        `pusion.__init__.py` file.

    """

    # List of supported PAC tuples (see above). This attribute is set by subclasses, i.e. combiner implementations.
    _SUPPORTED_PAC = []
    # List of all (combiner class, PAC)-mappings established during combiner registrations.
    _ALL_PAC_MAPPINGS = []
    # List of all combiner classes recognized by the framework.
    _ALL_COMBINERS = []

    # A shortname identifying the combiner.
    SHORT_NAME = None

    def __init__(self):
        self.pac = None
        self.coverage = None

    def __init_subclass__(cls, **kwargs):
        """
        Python-hook which registers all inheriting combiner classes defining `_SUPPORTED_PAC`.
        """
        super().__init_subclass__(**kwargs)
        if '_SUPPORTED_PAC' not in cls.__dict__:
            # A combiner class cannot be registered if the attribute _SUPPORTED_PAC is unset.
            return
        for pac in cls._SUPPORTED_PAC:
            cls._ALL_PAC_MAPPINGS.append((cls, pac))
        if cls not in cls._ALL_COMBINERS:
            cls._ALL_COMBINERS.append(cls)

    @classmethod
    def obtain(cls, config):
        """
        Obtain a combiner registered by the framework.

        :param config: :class:`pusion.model.configuration.Configuration`. User-defined configuration.
        :return: A `combiner` object.
        """
        # Search for basis combiners supporting the given problem, assignment and coverage type (pac).
        if (config.method, config.get_pac()) in cls._ALL_PAC_MAPPINGS:
            return cls.__prepare_method(config.method, config.get_pac())
        # Search for specialized combiners supporting the given problem, assignment and coverage type (pac).
        for subclass_combiner in config.method.__subclasses__():
            if (subclass_combiner, config.get_pac()) in cls._ALL_PAC_MAPPINGS:
                return cls.__prepare_method(subclass_combiner, config.get_pac())
        raise ValidationError("The given configuration is not supported by '{}'.".format(config.method.__name__))

    @classmethod
    def __prepare_method(cls, method, pac):
        combiner = method()
        combiner.pac = pac
        return combiner

    @abstractmethod
    def combine(self, decision_tensor):
        """
        Abstract method. Combine decision outputs by combiner's implementation.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of decision outputs by different classifiers per sample.
        :return: `numpy.array` of shape `(n_samples, n_classes)`. A matrix of class assignments which represents fused
                decisions obtained by combiner's implementation. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        pass

    def set_coverage(self, coverage):
        """
        Set the coverage for complementary-redundant decisions.

        :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a
                classifier, which is identified by the positional index of the respective list.
        """
        self.coverage = coverage


class UtilityBasedCombiner(Combiner):
    """
    A combiner of type :class:`UtilityBasedCombiner` fuses decisions solely based on the outputs of the ensemble
    classifiers. It does not take any further information or evidence about respective ensemble classifiers into
    account.
    """
    _SUPPORTED_PAC = []

    def __init__(self):
        super().__init__()

    @abstractmethod
    def combine(self, decision_tensor):
        """
        Abstract method. Combine decision outputs by combiner's implementation.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of decision outputs by different classifiers per sample.
        :return: `numpy.array` of shape `(n_samples, n_classes)`. A matrix of class assignments which represents fused
                decisions obtained by combiner's implementation. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        pass


class TrainableCombiner(Combiner):
    """
    A combiner of type :class:`TrainableCombiner` needs to be trained using decision outputs of the ensemble classifiers
    with true class assignments in order to combine decisions of unknown samples.
    """
    _SUPPORTED_PAC = []

    def __init__(self):
        super().__init__()

    @abstractmethod
    def combine(self, decision_tensor):
        """
        Abstract method. Combine decision outputs by combiner's implementation.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of decision outputs by different classifiers per sample.
        :return: `numpy.array` of shape `(n_samples, n_classes)`. A matrix of class assignments which represents fused
                decisions obtained by combiner's implementation. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        pass

    @abstractmethod
    def train(self, decision_tensor, true_assignments, **kwargs):
        """
        Abstract method. Train combiner's implementation using decision outputs an appropriate true assignments.
         
        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of decision outputs by different classifiers per sample.
        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of class assignments which are considered true for each sample during the training procedure.
        :param **kwargs: The `**kwargs` parameter may be used to use additional test data for the AutoFusion selection
                procedure.
        """
        pass


class EvidenceBasedCombiner(Combiner):
    """
    A combiner of type :class:`EvidenceBasedCombiner` takes an additional evidence into account while combining outputs
    of ensemble classifiers. Thus, it is able to empower better classifiers in order to obtain a fusion result with
    higher overall classification performance.
    """
    _SUPPORTED_PAC = []

    def __init__(self):
        super().__init__()

    @abstractmethod
    def combine(self, decision_tensor):
        """
        Abstract method. Combine decision outputs by combiner's implementation.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of decision outputs by different classifiers per sample.
        :return: `numpy.array` of shape `(n_samples, n_classes)`. A matrix of class assignments which represents fused
                decisions obtained by combiner's implementation. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        pass

    @abstractmethod
    def set_evidence(self, evidence):
        """
        Abstract method. Set the evidence for evidence-based combiner implementations.
        The evidence is given by confusion matrices calculated according to Kuncheva :footcite:`kuncheva2014combining`.

        .. footbibliography::

        :param evidence: `list` of `numpy.array` elements of shape `(n_classes, n_classes)`. Confusion matrices
                for each ensemble classifier.
        """
        pass



import multiprocessing as mp

from sklearn.datasets import make_classification, make_multilabel_classification
from sklearn.model_selection import train_test_split

from pusion.util.processes import p_fit
from pusion.util.transformer import *


def run_multiclass_ensemble(classifiers, x_train, y_train, x_valid, y_valid, x_test, y_test, continuous_out=False,
                            parallelize=True):
    """
    TODO comments.
    Generate random multiclass, crisp and redundant classification outputs (assignments) for the given ensemble of
    classifiers.

    :param classifiers: Classifiers used to generate classification outputs.
            These need to implement `fit` and `predict` methods according to classifiers provided by `sklearn`.
    :param parallelize: If `True`, all classifiers are trained in parallel. Otherwise they are trained in sequence.
    :param continuous_out: If `True`, class assignments in `y_ensemble_valid` and `y_ensemble_test` are given as
            probabilities. Default value is `False`.
    :return: `tuple` of:
            - `y_ensemble_valid`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a validation dataset.
            - `y_valid`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the validation.
            - `y_ensemble_test`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a test dataset.
            - `y_test`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the test.
    """

    if not (y_train.shape[1] == y_valid.shape[1] == y_test.shape[1]):
        raise TypeError("Different number of classes across train, valid and test.")

    n_classes = y_train.shape[1]

    # Convert to label vector to train sklearn classifiers (requires pusion commit 39a687 from master branch).
    y_train = class_assignment_matrix_to_label_vector(y_train)

    # Classifier training
    if parallelize:
        # Parallelized classifier training
        # Create a thread-safe queue for classifiers
        queue = mp.Manager().Queue()
        processes = []
        for i, clf in enumerate(classifiers):
            processes.append(mp.Process(target=p_fit, args=(i, clf, x_train, y_train, queue)))
        for p in processes:
            p.start()
        for p in processes:
            p.join()
        while not queue.empty():
            i, clf = queue.get()
            classifiers[i] = clf
    else:
        # Sequential classifier training
        for i, clf in enumerate(classifiers):
            print("Train classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            clf.fit(x_train, y_train)

    # Classifier validation to generate combiner training data
    y_ensemble_valid = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_valid.append(clf.predict_proba(x_valid))
    else:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_valid.append(clf.predict(x_valid))

    # Classifier test
    y_ensemble_test = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_test.append(clf.predict_proba(x_test))
    else:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_test.append(clf.predict(x_test))

    # Transform to matrix representation
    if not continuous_out:
        y_ensemble_valid = transform_label_tensor_to_class_assignment_tensor(y_ensemble_valid, n_classes)
        y_ensemble_test = transform_label_tensor_to_class_assignment_tensor(y_ensemble_test, n_classes)

    # Transform to numpy tensors if possible
    y_ensemble_valid = tensorize(y_ensemble_valid)
    y_ensemble_test = tensorize(y_ensemble_test)

    return y_ensemble_valid, y_valid, y_ensemble_test, y_test



import json
import ntpath
import shutil
from pathlib import Path

import pickle


def load_pickle_files_as_data(file_paths):
    """
    Load pickle files containing decision outputs as an data array.

    :param file_paths: A List of file paths to the individual pickle files.
    :return: A data array.
    """
    data = []
    for file_path in file_paths:
        with (open(file_path, "rb")) as handle:
            data.append(pickle5.load(handle))
    return data


def dump_pusion_data(data, file_path):
    """
    Dump classification output data to the given file using pickle.

    :param data: A data dictionary.
    :param file_path: Location of the output pickle file.
    """
    with open(file_path, "wb") as handle:
        pickle5.dump(data, handle, protocol=pickle5.HIGHEST_PROTOCOL)


def dump_data_as_txt(data, name, identifier):
    """
    Dump a data dictionary to the JSON file for a given evaluation unit.

    :param data: A data dictionary.
    :param name: The file name.
    :param identifier: The identifier of the current evaluation unit (e.g. date/time).
    """
    directory = "res/eval_" + identifier
    Path(directory).mkdir(parents=True, exist_ok=True)
    with open(directory + "/" + name + ".txt", 'w') as file:
        file.write(json.dumps(data, indent=4))


def save(plot_instance, name, identifier):
    """
    Save the plot instance for a given evaluation unit to the SVG and the PDF file, respectively.

    :param plot_instance: `matplotlib.pyplot`-instance.
    :param name: The file name.
    :param identifier: The identifier of the current evaluation unit (e.g. date/time).
    """
    directory = "res/eval_" + identifier
    Path(directory).mkdir(parents=True, exist_ok=True)
    plot_instance.savefig(directory + "/" + name + ".svg", bbox_inches="tight")
    plot_instance.savefig(directory + "/" + name + ".pdf", bbox_inches="tight")


def save_evaluator(file, identifier):
    """
    Save the evaluation script for a given evaluation unit.

    :param file: The Python file. (E.g. referenced by __file__).
    :param identifier: The identifier of the current evaluation unit (e.g. date/time).
    """
    directory = "res/eval_" + identifier
    Path(directory).mkdir(parents=True, exist_ok=True)
    shutil.copy(file, directory + "/" + ntpath.basename(file) + ".txt")



from pusion.util.constants import *


class Configuration:
    """
    The :class:`Configuration` forms the main parameter of the decision fusion framework. Based on this, the framework
    is able to check the compatibility of a respective decision fusion method to the given decision outputs.
    A configuration may be defined by the user or auto-detected by the framework.

    :param method: An explicit method provided by the framework which should be applied on the input data set.
            See `pusion.Method` for possible options.
    :param problem: Input problem type. See `pusion.util.constants.Problem` for possible options.
    :param assignment_type: The class assignment describes memberships to each individual class for a sample.
            A class assignment type is either crisp or continuous. Crisp assignments are equivalent to labels and
            continuous assignments represent probabilities for each class being true.
            See `pusion.util.constants.AssignmentType` for possible options.
    :param coverage_type: The classification coverage states for each input classifier, which classes it is able to
            decide. A classifier ensemble may yield a redundant, complementary or complementary-redundant coverage.
            See `pusion.util.constants.CoverageType` for possible options.
    """
    def __init__(self, method,
                 problem=Problem.GENERIC,
                 assignment_type=AssignmentType.GENERIC,
                 coverage_type=CoverageType.GENERIC):

        self.method = method
        self.problem = problem
        self.assignment_type = assignment_type
        self.coverage_type = coverage_type

    def get_tuple(self):
        """
        :return: A `tuple` of method, problem, assignment type and coverage type.
        """
        return self.method, self.problem, self.assignment_type, self.coverage_type

    def get_pac(self):
        """
        :return: A `tuple` of problem, assignment type and coverage type. This tuple is also referred to as PAC.
        """
        return self.problem, self.assignment_type, self.coverage_type



from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier

from pusion.core.combiner import TrainableCombiner
from pusion.util.transformer import *
from pusion.util.constants import *


class KNNCombiner(TrainableCombiner):
    """
    The :class:`KNNCombiner` (kNN) is a learning and classifier-based combiner that converts multiple decision
    outputs into new features, which in turn are used to train this combiner.
    The kNN combiner (k=5) uses uniform weights for all neighbors and the standard Euclidean metric for the distance.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'kNN'

    def __init__(self):
        TrainableCombiner.__init__(self)
        self.classifier = KNeighborsClassifier()

    def train(self, decision_tensor, true_assignments):
        """
        Train the kNN combiner by fitting the `k` nearest neighbors (k=5) model with given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported.
        This procedure transforms decision outputs into a new feature space.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        # transfer decisions into a new feature space
        featured_decisions = decision_profiles.reshape((decision_profiles.shape[0], -1))

        self.classifier.fit(featured_decisions, true_assignments)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by the `k` nearest neighbors (k=5) model.
        Both continuous and crisp classification outputs are supported. Combining requires a trained
        :class:`DecisionTreeCombiner`.
        This procedure transforms decision outputs into a new feature space.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by kNN. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        # transfer decisions into a new feature space
        featured_decisions = decision_profiles.reshape((decision_profiles.shape[0], -1))

        # return np.array(self.classifier.predict_proba(featured_decisions))[:, :, 1].T  # continuous output
        return self.classifier.predict(featured_decisions)


class CRKNNCombiner(KNNCombiner):
    """
    The :class:`CRKNNCombiner` is a modification of :class:`KNNCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as a constant, respectively. To use methods :meth:`train` and
    :meth:`combine` a coverage needs to be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
    ]

    def __init__(self):
        super().__init__()

    def train(self, decision_outputs, true_assignments):
        """
        Train the kNN combiner model by fitting the `k` nearest neighbors (k=5) model with given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported.
        This procedure transforms decision outputs into a new feature space.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains either crisp or continuous
                decision outputs per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        super().train(t_decision_outputs, true_assignments)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by the `k` nearest neighbors (k=5) model.
        Both continuous and crisp classification outputs are supported. Combining requires a trained
        :class:`DecisionTreeCombiner`.
        This procedure transforms decision outputs into a new feature space.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of crisp or continuous class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.negative(np.ones((n_classifiers, n_decisions, n_classes)))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



import numpy as np

from pusion.core.combiner import UtilityBasedCombiner
from pusion.util.transformer import decision_tensor_to_decision_profiles
from pusion.util.constants import *


class MacroMajorityVoteCombiner(UtilityBasedCombiner):
    """
    The :class:`MacroMajorityVoteCombiner` (MAMV) is based on a variation of the general majority vote method.
    The fusion consists of a decision vector which is given by the majority of the classifiers in the ensemble for a
    sample. MAMV does not consider outputs for each individual class (macro).
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'MAMV'

    def __init__(self):
        UtilityBasedCombiner.__init__(self)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by majority voting across all classifiers considering the most common classification
        assignment (macro). Only crisp classification outputs are supported.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of crisp decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments obtained by MAMV. Axis 0 represents samples and
                axis 1 the class labels which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        fused_decisions = np.zeros_like(decision_tensor[0])
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        for i in range(len(decision_profiles)):
            dp = decision_profiles[i]
            unique_decisions = np.unique(dp, axis=0, return_counts=True)
            decisions = unique_decisions[0]
            counts = unique_decisions[1]
            fused_decisions[i] = decisions[np.argmax(counts)]
        return fused_decisions



import math

import numpy as np
from sklearn.metrics import confusion_matrix, multilabel_confusion_matrix


def confusion_matrices_to_accuracy_vector(confusion_matrix_tensor):
    """
    Convert confusion matrices of respective classification to an accuracy vector.

    :param confusion_matrix_tensor: `numpy.array` of shape `(n_classifiers, n_classes, n_classes)`
            Confusion matrices.
    :return: One-dimensional `numpy.array` of shape of length `n_classifiers` containing the accuracy for each confusion
            matrix.
    """
    accuracies = np.zeros(len(confusion_matrix_tensor))
    for i in range(len(confusion_matrix_tensor)):
        accuracies[i] = confusion_matrix_to_accuracy(confusion_matrix_tensor[i])
    return accuracies


def confusion_matrix_to_accuracy(cm):
    """
    Calculate the accuracy out of the given confusion matrix.

    :param cm: `numpy.array` of shape `(n_classes, n_classes)`.
            Confusion matrix.
    :return: The accuracy.
    """
    cm = np.array(cm)
    return np.sum(np.diagonal(cm)) / np.sum(cm)


def multilabel_cr_confusion_matrices_to_avg_accuracy(label_cms):
    """
    Calculate the average accuracy for the given confusion matrices generated from complementary redundant multilabel
    output.

    :param label_cms: `list` of confusion matrices given as 2-dimensional `numpy.array` respectively.
    :return: The average accuracy.
    """
    avg = 0.0
    for label_cm in label_cms:
        avg += confusion_matrix_to_accuracy(label_cm)
    return avg / len(label_cms)


def decision_tensor_to_decision_profiles(decision_tensor):
    """
    Transform the given decision tensor to decision profiles for each respective sample.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of either crisp or continuous decision outputs by different classifiers per sample.
    :return: `numpy.array` of shape `(n_samples, n_classifiers, n_classes)`.
            Decision profiles.
    """
    return np.array(decision_tensor).transpose((1, 0, 2))


def multilabel_predictions_to_decisions(predictions, threshold=0.5):
    """
    Transform a continuously valued tensor of multilabel decisions to crisp decision outputs.

    :param predictions: `numpy.array` of any shape. Continuous predictions.
    :param threshold: `float`. A threshold value, based on which the crisp output is constructed.
    :return: `numpy.array` of the same shape as ``predictions``. Crisp decision outputs.
    """
    return (predictions >= threshold) * np.ones_like(predictions)


def multiclass_predictions_to_decisions(predictions):
    """
    Transform a continuously valued matrix of multiclass decisions to crisp decision outputs.

    :param predictions: `numpy.array` of shape `(n_samples, n_classes)`. Continuous predictions.
    :return: `numpy.array` of the same shape as ``predictions``. Crisp decision outputs.
    """
    if predictions.ndim == 1:
        return predictions
    decisions = np.zeros_like(predictions)
    decisions[np.arange(len(decisions)), predictions.argmax(axis=1)] = 1
    return decisions


def multilabel_prediction_tensor_to_decision_tensor(predictions):
    """
    Transform a continuously valued tensor of multilabel decisions to crisp decision outputs.

    :param predictions: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`. Continuous predictions.
    :return: `numpy.array` of the same shape as ``predictions``. Crisp decision outputs.
    """
    # decision_tensor = np.zeros_like(predictions)
    decision_outputs = []
    for i, pred_matrix in enumerate(predictions):
        # decision_tensor[i] = multiclass_predictions_to_decisions(pred_matrix)
        decision_outputs.append(multilabel_predictions_to_decisions(pred_matrix))
    return tensorize(decision_outputs)


def multiclass_prediction_tensor_to_decision_tensor(predictions):
    """
    Transform a continuously valued tensor of multiclass decisions to crisp decision outputs.

    :param predictions: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`. Continuous predictions.
    :return: `numpy.array` of the same shape as ``predictions``. Crisp decision outputs.
    """
    # decision_tensor = np.zeros_like(predictions)
    decision_outputs = []
    for i, pred_matrix in enumerate(predictions):
        # decision_tensor[i] = multiclass_predictions_to_decisions(pred_matrix)
        decision_outputs.append(multiclass_predictions_to_decisions(pred_matrix))
    return tensorize(decision_outputs)


def decision_tensor_to_configs(decision_outputs):
    """
    Transform decision outputs to decision configs. A decision config shows concatenated classification outputs of each
    classifier per sample.

    :param decision_outputs: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
            `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
            due to the coverage.
    :return: `numpy.array` of shape `(n_samples, n_classes*)`, `n_classes*` is the sum of all classes covered by
            all classifiers.
    """
    return np.concatenate(decision_outputs, axis=1)


def multiclass_assignments_to_labels(assignments):
    """
    Transform multiclass assignments to labels. A matrix of shape `(n_samples, n_classes)` is converted to a vector
    of shape `(n_samples,)`, with element-wise labels represented in integers from `0` to `n_classes - 1`.

    :param assignments: `numpy.array` of shape `(n_samples, n_classes)`. Multiclass assignments.
    :return: `numpy.array` of shape `(n_samples,)` with an integer label per element.
    """
    assignments = np.array(assignments)
    return np.argmax(assignments, axis=assignments.ndim - 1)


def transform_label_tensor_to_class_assignment_tensor(label_tensor, n_classes):
    """
    Transform a label tensor of shape `(n_classifiers, n_samples)` to the tensor of class assignments of shape
    `(n_classifiers, n_samples, n_classes)`. A label is an integer between `0` and `n_classes - 1`.

    :param label_tensor: `numpy.array` of shape `(n_classifiers, n_samples)`. Label tensor.
    :param n_classes: Number of classes to be considered.
    :return: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`. Class assignment tensor (decision tensor).
    """
    label_tensor = np.array(label_tensor)
    assignments = np.zeros((label_tensor.shape[0], label_tensor.shape[1], n_classes))
    for i in range(label_tensor.shape[0]):
        assignments[i, np.arange(label_tensor.shape[1]), label_tensor[i]] = 1
    return assignments


def transform_label_vector_to_class_assignment_matrix(label_vector, n_classes=None):
    """
    Transform labels to multiclass assignments. A vector of shape `(n_samples,)`, with element-wise labels is converted
    to the assignment matrix of shape `(n_samples, n_classes)`.

    :param label_vector: `numpy.array` of shape `(n_samples,)` with an integer label per element.
    :param n_classes: Number of classes to be considered.
    :return: `numpy.array` of shape `(n_samples, n_classes)`. Multiclass assignments.
    """
    label_vector = np.array(label_vector, dtype=int)
    if n_classes is None:
        n_classes = np.max(label_vector) + 1

    assignments = np.zeros((len(label_vector), n_classes), dtype=int)
    assignments[np.arange(len(label_vector)), label_vector] = 1
    return assignments


def class_assignment_tensor_to_label_tensor(class_assignment_tensor):
    """
    Transform multiclass class assignments into a label tensor. A tensor of shape
    `(n_classifiers, n_samples, n_classes)` is converted into a tensor of shape `(n_classifiers, n_samples)` holding the
    class identifier (label) in each element.

    :param class_assignment_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Class assignment tensor (decision tensor).
    :return: `numpy.array` of shape `(n_classifiers, n_samples)`. Label tensor.
    """
    class_assignment_tensor = np.array(class_assignment_tensor).astype(int)
    return np.argmax(class_assignment_tensor == 1, axis=2)


def class_assignment_matrix_to_label_vector(class_assignment_matrix):
    """
    Transform multiclass class assignments into a label vector. A matrix of shape
    `(n_samples, n_classes)` is converted into a vector of shape `(n_samples,)` holding the class identifier (label)
    in each element.

    :param class_assignment_matrix: `numpy.array` of shape `(n_samples, n_classes)`. Class assignment matrix
            (decision matrix).
    :return: `numpy.array` of shape `(n_samples,)`. Label vector.
    """
    class_assignment_matrix = np.array(class_assignment_matrix).astype(int)
    return np.argmax(class_assignment_matrix == 1, axis=1)


def multilabel_to_multiclass_assignments(decision_tensor):
    """
    Transform the multilabel decision tensor to the equivalent multiclass decision tensor using the power set method.
    The multilabel class assignments are considered as a binary number which represents a new class in the
    multiclass decision space. E.g. the assignment to the classes `0` and `2` (`[1,0,1]`) is converted to the class `5`,
    which is one of the `2^3` classes in the multiclass decision space.
    This method is inverse to the ``multiclass_to_multilabel_assignments`` method.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multilabel decision outputs by different classifiers per sample.
    :return: `numpy.array` of shape `(n_classifiers, n_samples, 2^n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    """
    decision_tensor = np.array(decision_tensor).astype(int)
    input_is_matrix = False
    if decision_tensor.ndim == 2:
        decision_tensor = decision_tensor[np.newaxis, :]
        input_is_matrix = True
    mc_decision_tensor = np.zeros((decision_tensor.shape[0],
                                   decision_tensor.shape[1],
                                   2 ** decision_tensor.shape[2]), dtype=int)
    for i in range(decision_tensor.shape[0]):
        for j in range(decision_tensor.shape[1]):
            new_class_index = sum(b << i for i, b in enumerate(decision_tensor[i, j]))
            mc_decision_tensor[i, j, new_class_index] = 1
    if input_is_matrix:
        return mc_decision_tensor.squeeze()
    return mc_decision_tensor


def multiclass_to_multilabel_assignments(decision_tensor):
    """
    Transform the multiclass decision tensor to the equivalent multilabel decision tensor using the inverse
    power set method. The multiclass assignment is considered as a decimal which is converted to a binary number, which
    in turn represents the multilabel class assignment. E.g. the class assignment to the class `3` `([0,0,0,1])` is
    converted to the multilabel class assignment `[1,1]` (classes `0` and `1` in the multilabel decision space).
    This method is inverse to the ``multilabel_to_multiclass_assignments`` method.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multilabel decision outputs by different classifiers per sample.
    :return: `numpy.array` of shape `(n_classifiers, n_samples, log_2(n_classes))`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    """
    decision_tensor = np.array(decision_tensor)
    input_is_matrix = False
    if decision_tensor.ndim == 2:
        decision_tensor = decision_tensor[np.newaxis, :]
        input_is_matrix = True
    ml_decision_tensor = np.zeros((decision_tensor.shape[0],
                                   decision_tensor.shape[1],
                                   int(math.log(decision_tensor.shape[2], 2))), dtype=int)
    for i in range(decision_tensor.shape[0]):
        for j in range(decision_tensor.shape[1]):
            class_indices = np.where(decision_tensor[i, j])[0]
            if class_indices.size > 0:
                ml_decision = np.flip(np.array([int(x) for x in bin(class_indices[0])[2:]]))
                ml_decision_tensor[i, j, range(len(ml_decision))] = ml_decision
    if input_is_matrix:
        return ml_decision_tensor.squeeze()
    return ml_decision_tensor


def tensorize(decision_outputs):
    """
    Convert `list` decision outputs to `numpy.array` decision tensor, if possible.
    """
    try:
        return np.array(decision_outputs, dtype=float)
    except ValueError:
        return decision_outputs


def intercept_normal_class(y, override=False):
    """
    Intercept the normal class for the given decision matrix, i.e. a normal class is assigned to each zero vector
    class assignment. E.g. the assignment `[0,0,0,0]` is transformed to `[1,0,0,0]`, under the assumption that `0`
    is a normal class.

    :param y: `numpy.array` of shape `(n_samples, n_classes)`. Matrix of decision outputs.
    :param override: If `true`, the class `0` is assumed as a normal class. Otherwise a new class is prepended to
            existing classes.
    :return: `numpy.array` of shape `(n_samples, n_classes)` for `override=False`.
            `numpy.array` of shape `(n_samples, n_classes + 1)` for `override=True`.
            Matrix of decision outputs with intercepted normal class.
    """
    y = np.array(y)
    if not override:
        y = np.column_stack((np.zeros(y.shape[0], dtype=int), y))
    z_indices = np.where(~y.any(axis=1))[0]
    # set the normal class to 1 for zero vectors
    y[z_indices, 0] = 1
    return y


def intercept_normal_class_in_tensor(decision_tensor, override=False):
    """
    Intercept the normal class for the given decision matrix, i.e. a normal class is assigned to each zero vector
    class assignment. E.g. the assignment `[0,0,0,0]` is transformed to `[1,0,0,0]`, under the assumption that `0`
    is a normal class.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of decision outputs by different classifiers per sample.
    :param override: If `true`, the class `0` is assumed as a normal class. Otherwise a new class is prepended to
            existing classes.
    :return: `numpy.array` of shape `(n_samples, n_classes)` for `override=False`.
            `numpy.array` of shape `(n_samples, n_classes + 1)` for `override=True`.
            Matrix of decision outputs with intercepted normal class.
        """
    normalized_decision_outputs = []
    for i in range(len(decision_tensor)):
        normalized_decision_outputs.append(intercept_normal_class(decision_tensor[i], override))
    return tensorize(normalized_decision_outputs)



import numpy as np

from pusion.core.combiner import UtilityBasedCombiner
from pusion.util.transformer import multilabel_predictions_to_decisions
from pusion.util.constants import *


class SimpleAverageCombiner(UtilityBasedCombiner):
    """
    The :class:`SimpleAverageCombiner` (AVG) fuses decisions using the arithmetic mean rule.
    The mean is calculated between decision vectors obtained by multiple ensemble classifiers for a sample.
    The AVG combiner is unaware of the input problem (multiclass/multilabel) or the assignment type (crisp/continuous).
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'AVG'

    def __init__(self):
        UtilityBasedCombiner.__init__(self)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by averaging the class support of each classifier in the given ensemble.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp assignments which represents fused
                decisions obtained by the AVG method. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        # return np.mean(decision_tensor, axis=0)  # continuous output
        return multilabel_predictions_to_decisions(np.mean(decision_tensor, axis=0), .5)


class CRSimpleAverageCombiner(SimpleAverageCombiner):
    """
    The :class:`CRSimpleAverageCombiner` is a modification of :class:`SimpleAverageCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed to a unified
    tensor representation supporting undefined class assignments. The mean is calculated only for assignments which
    are defined. To call :meth:`combine` a coverage needs to be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
    ]

    def __init__(self):
        super().__init__()

    def combine(self, decision_outputs):
        """
        Combine decision outputs by averaging the defined class support of each classifier in the given ensemble.
        Undefined class supports are excluded from averaging.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains either crisp or continuous
                decision outputs per sample.

        :return: A matrix (`numpy.array`) of crisp assignments which represents fused
                decisions obtained by the AVG method. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        # return np.nanmean(t_decision_outputs, axis=0)  # continuous output
        return multilabel_predictions_to_decisions(np.nanmean(t_decision_outputs, axis=0), .5)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.full((n_classifiers, n_decisions, n_classes), np.nan)
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



import multiprocessing as mp

from pusion.auto.detector import *
from pusion.core.combiner import *
from pusion.util.processes import *


class GenericCombiner(TrainableCombiner, EvidenceBasedCombiner, UtilityBasedCombiner):
    """
    The `GenericCombiner` (GC) allows for automatic decision fusion using all methods provided by the framework, which
    are applicable to the given problem. The key feature of this combiner is the transparency in terms of it's outer
    behaviour. Based on the usage (i.e. method calls) and the automatically detected configuration,
    the `GenericCombiner` preselects all compatible methods from `pusion.core`. The main purpose is to retrieve fusion
    results obtained by the all applicable methods. The main difference to the `AutoCombiner` is that decision fusion
    results are handed over to the user for further comparison and selection. Thus, GC is not suitable for
    the online fusion.
    """
    _SUPPORTED_PAC = [
        (Problem.GENERIC, AssignmentType.GENERIC, CoverageType.GENERIC),
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT)
    ]

    def __init__(self):
        super().__init__()
        self.problem = Problem.GENERIC
        self.assignment_type = AssignmentType.GENERIC
        self.coverage_type = CoverageType.REDUNDANT

        self.combiner_type_selection = []
        self.combiners = []
        self.multi_combiner_decision_tensor = []

        self.multi_combiner_train_runtimes = []
        self.multi_combiner_combine_runtimes = []

        self.evidence = None
        self.parallel = True

    def set_coverage(self, coverage):
        """
        Set the coverage in case of complementary-redundant classification data.

        :param coverage: The coverage is described by using a nested list. Each list describes the classifier based on
                its position. Elements of those lists (integers) describe the actual class coverage of the respective
                classifier. E.g., with ``[[0,1], [0,2,3]]`` the classes 0,1 are covered by the first classifier and
                0,2,3 are covered by the second one.
        """
        self.coverage = coverage
        self.coverage_type = determine_coverage_type(coverage)

    def set_evidence(self, evidence):
        """
        Set the evidence for evidence based combiners. This method preselects all combiners of type
        `EvidenceBasedCombiner`.

        :param evidence: `numpy.array` of shape `(n_classifiers, n_classes, n_classes)`.
                Confusion matrices for each of `n` classifiers.
        """
        self.__add_combiner_type(EvidenceBasedCombiner)
        self.evidence = evidence

    def train(self, decision_tensor, true_assignments):
        """
        Train the Generic Combiner. This method detects the configuration based on the ``decision_tensor`` and
        trains all trainable combiners that are applicable to this configuration.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        if self.parallel:
            self.train_par(decision_tensor, true_assignments)
        else:
            self.train_seq(decision_tensor, true_assignments)

    def train_par(self, decision_tensor, true_assignments):
        """
        Train the Generic Combiner by training individual combiners in parallel.
        This method detects the configuration based on the ``decision_tensor`` and trains all trainable combiners
        that are applicable to this configuration.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        self.__add_combiner_type(UtilityBasedCombiner)
        self.__add_combiner_type(TrainableCombiner)
        self.problem = determine_problem(decision_tensor)
        self.assignment_type = determine_assignment_type(decision_tensor)
        self.combiners = self.__obtain_from_registered_methods(self.get_pac(), self.combiner_type_selection)
        self.__prepare()

        # Create a thread-safe queue for combiners
        queue = mp.Manager().Queue()
        processes = []
        for i, combiner in enumerate(self.combiners):
            if isinstance(combiner, TrainableCombiner):
                process = mp.Process(target=p_train, args=(i, combiner, decision_tensor, true_assignments, queue))
                processes.append(process)
            else:
                queue.put((i, combiner, np.nan))
        for p in processes:
            p.start()
        for p in processes:
            p.join()

        self.combiners = [None for _ in range(len(self.combiners))]
        while not queue.empty():
            i, combiner, t_elapsed = queue.get()
            self.combiners[i] = combiner
            self.multi_combiner_train_runtimes.append((i, t_elapsed))

    def train_seq(self, decision_tensor, true_assignments):
        """
        Train the Generic Combiner by training individual combiners in sequence.
        This method detects the configuration based on the ``decision_tensor`` and trains all trainable combiners
        that are applicable to this configuration.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        self.__add_combiner_type(UtilityBasedCombiner)
        self.__add_combiner_type(TrainableCombiner)
        self.problem = determine_problem(decision_tensor)
        self.assignment_type = determine_assignment_type(decision_tensor)
        self.combiners = self.__obtain_from_registered_methods(self.get_pac(), self.combiner_type_selection)
        self.__prepare()

        self.multi_combiner_train_runtimes = []
        for i, combiner in enumerate(self.combiners):
            if isinstance(combiner, TrainableCombiner):
                t_begin = time.perf_counter()
                combiner.train(decision_tensor, true_assignments)
                t_elapsed = time.perf_counter() - t_begin
                self.multi_combiner_train_runtimes.append((i, t_elapsed))

    def combine(self, decision_tensor):
        """
        Combine decision outputs using the AutoCombiner (AC) model. Both continuous and crisp classification outputs are
        supported. This procedure involves combining decision outputs by each individual method which is applicable
        to the detected configuration.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: list of `numpy.array` of shape `(n_samples, n_classifiers)`.
                Fusion results obtained by selected fusion methods.
                The list is aligned with the list of preselected fusion methods (retrievable by ``get_combiners()``).
        """
        if self.parallel:
            return self.combine_par(decision_tensor)
        else:
            return self.combine_seq(decision_tensor)

    def combine_par(self, decision_tensor):
        """
        Combine decision outputs by GC. Both continuous and crisp classification outputs are supported.
        This procedure involves combining decision outputs by each individual method which is applicable
        to the detected configuration. Each combine procedure is spawned in a separate thread and thus performed
        in parallel.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: list of `numpy.array` of shape `(n_samples, n_classifiers)`.
                Fusion results obtained by selected fusion methods.
                The list is aligned with the list of preselected fusion methods (retrievable by ``get_combiners()``).
        """
        self.__add_combiner_type(UtilityBasedCombiner)
        self.problem = determine_problem(decision_tensor)
        self.assignment_type = determine_assignment_type(decision_tensor)

        if not self.__combiner_type_selected(TrainableCombiner):
            self.combiners = self.__obtain_from_registered_methods(self.get_pac(), self.combiner_type_selection)

        self.__prepare()

        # Create a thread-safe queue for combiners
        queue = mp.Manager().Queue()
        processes = []
        for i, combiner in enumerate(self.combiners):
            process = mp.Process(target=p_combine, args=(i, combiner, decision_tensor, queue))
            processes.append(process)
        for p in processes:
            p.start()
        for p in processes:
            p.join()

        self.multi_combiner_decision_tensor = [None for _ in range(len(self.combiners))]
        while not queue.empty():
            i, decision_matrix, t_elapsed = queue.get()
            self.multi_combiner_decision_tensor[i] = decision_matrix
            self.multi_combiner_combine_runtimes.append((i, t_elapsed))
        return self.multi_combiner_decision_tensor

    def combine_seq(self, decision_tensor):
        """
        Combine decision outputs by GC. Both continuous and crisp classification outputs are supported.
        This procedure involves combining decision outputs by each individual method which is applicable
        to the detected configuration. Each combine procedure is initiated in sequence.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: list of `numpy.array` of shape `(n_samples, n_classifiers)`.
                Fusion results obtained by selected fusion methods.
                The list is aligned with the list of preselected fusion methods (retrievable by ``get_combiners()``).
        """
        self.__add_combiner_type(UtilityBasedCombiner)
        self.problem = determine_problem(decision_tensor)
        self.assignment_type = determine_assignment_type(decision_tensor)

        if not self.__combiner_type_selected(TrainableCombiner):
            self.combiners = self.__obtain_from_registered_methods(self.get_pac(), self.combiner_type_selection)

        self.__prepare()

        self.multi_combiner_decision_tensor = []
        self.multi_combiner_combine_runtimes = []
        for i, combiner in enumerate(self.combiners):
            t_begin = time.perf_counter()
            self.multi_combiner_decision_tensor.append(combiner.combine(decision_tensor))
            t_elapsed = time.perf_counter() - t_begin
            self.multi_combiner_combine_runtimes.append((i, t_elapsed))
        return self.multi_combiner_decision_tensor

    def __prepare(self):
        for combiner in self.combiners:
            if isinstance(combiner, EvidenceBasedCombiner) and self.evidence:
                combiner.set_evidence(self.evidence)
            if self.coverage_type == CoverageType.COMPLEMENTARY \
                    or self.coverage_type == CoverageType.COMPLEMENTARY_REDUNDANT:
                combiner.set_coverage(self.coverage)

    def __obtain_from_registered_methods(self, pac, combiner_types):
        methods = []
        for combiner_type in combiner_types:
            for map_entry in super()._ALL_PAC_MAPPINGS:
                method, pac_ = map_entry[0], map_entry[1]
                if pac_ == pac and method not in methods and not issubclass(method, GenericCombiner) \
                        and method is not GenericCombiner and issubclass(method, combiner_type):
                    methods.append(method)
        return [method() for method in methods]

    def __add_combiner_type(self, combiner_type):
        if combiner_type not in self.combiner_type_selection:
            self.combiner_type_selection.append(combiner_type)

    def __combiner_type_selected(self, combiner_type):
        return combiner_type in self.combiner_type_selection

    def get_pac(self):
        """
        :return: `tuple` of detected problem, assignment type and coverage type.
        """
        return self.problem, self.assignment_type, self.coverage_type

    def get_combiners(self):
        """
        :return: list of core methods preselected by the `AutoCombiner`.
        """
        return self.combiners

    def get_combiner_type_selection(self):
        """
        :return: list of combiner types established by usage.
        """
        return self.combiner_type_selection

    def get_multi_combiner_decision_tensor(self):
        """
        :return: list of `numpy.array` of shape `(n_samples, n_classifiers)`.
                Fusion results obtained by selected fusion methods.
                The list is aligned with the list of preselected fusion methods (retrievable by ``get_combiners()``).
        """
        return self.multi_combiner_decision_tensor

    def get_multi_combiner_runtimes(self):
        """
        :return: A `tuple` of two lists of tuples describing the train and combine runtimes respectively.
                Each inner tuple key value indexes the list of preselected fusion methods
                (retrievable by ``get_combiners()``).
        """
        return self.multi_combiner_train_runtimes, self.multi_combiner_combine_runtimes

    def set_parallel(self, parallel=True):
        """
        Set whether the training and the combining of selected combiners should be executed sequentially or in parallel.
        :param parallel: If `True`, training and combining is performed in parallel respectively. Otherwise in sequence.
        """
        self.parallel = parallel



import numpy as np

from pusion.model.report import Report
from pusion.util.generator import *


class Evaluation:
    """
    :class:`Evaluation` provides methods for evaluating decision outputs (i.e. combiners and classifiers) with different
    problems and coverage types.

    :param argv: Performance metric functions.
    """
    def __init__(self, *argv):
        self.metrics = []
        self.instances = None
        self.performance_matrix = None
        self.runtime_matrix = None
        self.set_metrics(*argv)

    def evaluate(self, true_assignments, decision_tensor):
        """
        Evaluate the decision outputs with already set classification performance metrics.

        .. warning::

            This evaluation is only applicable on redundant multiclass or multilabel decision outputs.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which are considered true for the evaluation.
        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of crisp decision outputs by different classifiers per sample.
        :return: `numpy.array` of shape `(n_instances, n_metrics)`. Performance matrix containing performance values
                for each set instance row-wise and each set performance metric column-wise.
        """
        self.__check()
        # generalize to decision_tensor if a matrix is given
        decision_tensor = np.array(decision_tensor)
        if decision_tensor.ndim < 3:
            decision_tensor = np.expand_dims(decision_tensor, axis=0)

        if len(self.instances) != len(decision_tensor):
            raise TypeError("`decision_tensor` is not aligned with the number of instances.")

        performance_matrix = np.full((len(decision_tensor), len(self.metrics)), np.nan)
        for i in range(len(decision_tensor)):
            for j in range(len(self.metrics)):
                metric = self.metrics[j]
                score = metric(true_assignments, decision_tensor[i])
                performance_matrix[i, j] = score
        self.performance_matrix = performance_matrix
        return performance_matrix

    def evaluate_cr_decision_outputs(self, true_assignments, decision_outputs, coverage=None):
        """
        Evaluate complementary-redundant decision outputs with already set classification performance metrics.
        The outputs of each classifier for each class is considered as a binary output and thus, the performance is
        calculated class-wise and averaged across all classes, which are covered by individual classifiers.

        .. note::

            This evaluation is applicable on complementary-redundant ensemble classifier outputs.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which are considered true for the evaluation.
        :param decision_outputs: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.
        :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a
                classifier, which is identified by the positional index of the respective list.
                If none set, the coverage for fully redundant classification is chosen by default.
        :return: `numpy.array` of shape `(n_instances, n_metrics)`. Performance matrix containing performance values
                for each set instance row-wise and each set performance metric column-wise.
        """
        self.__check()
        # generalize to 3-dimensional decision_outputs if a matrix is given
        if type(decision_outputs) == np.ndarray and decision_outputs.ndim < 3:
            decision_outputs = np.expand_dims(decision_outputs, axis=0)

        if coverage is None:
            coverage = [np.arange(true_assignments.shape[1], dtype=int) for _ in range(len(decision_outputs))]

        perf_matrix = np.full((1, len(self.metrics)), np.nan)
        for i, metric in enumerate(self.metrics):
            class_wise_mean_score = self.class_wise_mean_score(true_assignments, decision_outputs, coverage, metric)
            perf_matrix[0, i] = np.mean(class_wise_mean_score)

        self.performance_matrix = perf_matrix
        return perf_matrix

    def evaluate_cr_multi_combiner_decision_outputs(self, true_assignments, decision_tensor):
        """
        Evaluate decision outputs of multiple CR combiners with already set classification performance metrics.
        The evaluation is performed by :func:`evaluate_cr_decision_outputs` for each combiner.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which are considered true for the evaluation.
        :param decision_tensor: `numpy.array` of shape `(n_combiners, n_samples, n_classes)`.
                Tensor of crisp decision outputs by different combiners per sample.
        :return: `numpy.array` of shape `(n_instances, n_metrics)`. Performance matrix containing performance values
                for each set instance row-wise and each set performance metric column-wise.
        """
        self.__check()
        if len(self.instances) != len(decision_tensor):
            raise TypeError("`decision_tensor` is not aligned with the number of instances.")

        performance_matrix = np.full((len(decision_tensor), len(self.metrics)), np.nan)
        for i in range(len(decision_tensor)):
            dt = np.expand_dims(decision_tensor[i], axis=0)
            pm = self.evaluate_cr_decision_outputs(true_assignments, dt)
            performance_matrix[i] = np.squeeze(pm)
        self.performance_matrix = performance_matrix
        return performance_matrix

    def class_wise_mean_score(self, true_assignments, decision_outputs, coverage, metric):
        """
        Calculate the class-wise mean score with the given metric for the given classification outputs.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which are considered true for the evaluation.
        :param decision_outputs: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.
        :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a
                classifier, which is identified by the positional index of the respective list.
                If none set, the coverage for fully redundant classification is chosen by default.
        :param metric: The score metric.
        :return: `numpy.array` of shape `(n_classes,)`. The mean score per class across all classifiers.
        """
        self.__check()
        if len(decision_outputs) != len(coverage):
            raise TypeError("`decision_outputs` is not aligned to `coverage`.")

        true_assignments_per_coverage = [intercept_normal_class(true_assignments[:, cov], True) for cov in coverage]
        classifier_class_score_matrix = np.full((len(decision_outputs), true_assignments.shape[1]), np.nan)
        for i in range(len(decision_outputs)):
            for ci, j in enumerate(coverage[i]):
                decision_matrix = np.array(decision_outputs[i])
                pred_decision_vector = decision_matrix[:, ci]
                true_decision_vector = true_assignments_per_coverage[i][:, ci]
                classifier_class_score_matrix[i, j] = metric(true_decision_vector, pred_decision_vector)

        return np.nanmean(classifier_class_score_matrix, axis=0)

    def get_report(self):
        """
        :return: A summary `Report` of performed evaluations including all involved instances and performance metrics.
        """
        return Report(np.around(self.performance_matrix, 4), self.instances, self.metrics)

    def get_runtime_report(self):
        """
        :return: A summary `Report` of train and combine runtimes for all involved instances.
        """
        if self.runtime_matrix is not None:
            return Report(np.around(self.runtime_matrix, 4), self.instances, ['t_train (sec)', 't_comb (sec)'])
        raise TypeError("A runtime matrix is not set.")

    def get_instances(self):
        """
        :return: A `list` of instances (i.e. combiner or classifiers) been evaluated.
        """
        return self.instances

    def get_metrics(self):
        """
        :return: A `list` of performance metrics been used for evaluation.
        """
        return self.metrics

    def get_performance_matrix(self):
        """
        :return: `numpy.array` of shape `(n_instances, n_metrics)`. Performance matrix containing performance values
                for each set instance row-wise and each set performance metric column-wise.
        """
        return self.performance_matrix

    def get_runtime_matrix(self):
        """
        :return: `numpy.array` of shape `(n_instances, 2)`. Runtime matrix containing runtimes
                for each set instance row-wise. The column at index `0` describes train times and the column at index
                `1` describes combine times.
        """
        return self.runtime_matrix

    def get_top_n_instances(self, n=None, metric=None):
        """
        Retrieve top `n` best instances according to the given `metric` in a sorted order.

        :param n: `integer`. Number of instances to be retrieved. If unset, all instances are retrieved.
        :param metric: The metric all instances are sorted by. If unset, the first metric is used.
        :return: Evaluated top `n` instances.
        """
        self.__check()
        if self.performance_matrix is None:
            raise TypeError("No evaluation performed.")
        # set default parameter values
        if n is None:
            n = self.performance_matrix.shape[0]
        if metric is None:
            metric = self.metrics[0]

        metric_index = self.metrics.index(metric)
        performance_values = self.performance_matrix[:, metric_index]
        top_n_instance_indices = performance_values.argsort()[-n:][::-1]
        return [(self.instances[i], self.performance_matrix[i, metric_index]) for i in top_n_instance_indices]

    def get_top_instances(self, metric=None):
        """
        Retrieve best performing instances according to the given `metric`.
        Multiple instances may be returned having the identical best performance score.

        :param metric: The metric all instances were evaluated with. If unset, the first metric is used.
        :return: Evaluated top instances according to their performance.
        """
        top_n_instances = self.get_top_n_instances(metric=metric)
        top_score = None
        top_instances = []
        for i, t in enumerate(top_n_instances):
            if top_score is None:
                top_score = t[1]
                top_instances.append(t)
                continue
            if t[1] == top_instances[0][1]:
                top_instances.append(t)
        return top_instances

    def get_instance_performance_tuples(self, metric=None):
        """
        Retrieve (instance, performance) tuples created for to the given `metric`.

        :param metric: The metric all instances are evaluated by. If unset, the first set metric is used.
        :return: `list` of (instance, performance) tuples.
        """
        self.__check()
        if self.performance_matrix is None:
            raise TypeError("No evaluation performed.")
        # set default parameter values
        if metric is None:
            metric = self.metrics[0]

        metric_index = self.metrics.index(metric)
        return [(self.instances[i], self.performance_matrix[i, metric_index]) for i in range(len(self.instances))]

    def set_metrics(self, *argv):
        """
        :param argv: Performance metric functions.
        """
        self.metrics = []
        for metric in argv:
            if metric not in self.metrics:
                self.metrics.append(metric)

    def set_instances(self, instances):
        """
        :param instances: An instance or a `list` of instances to be evaluated, e.g. classifiers or combiners.
        """
        if type(instances) != list:
            instances = [instances]
        self.instances = instances

    def set_runtimes(self, runtimes):
        """
        :param runtimes: A `tuple` of two lists of tuples describing the train and combine runtimes respectively.
                Each runtime list is aligned with the list of set instances.
        """
        runtime_matrix = np.full((len(self.instances), 2), np.nan)
        for t in range(len(runtimes)):
            for i in range(len(runtimes[t])):
                runtime_matrix[runtimes[t][i][0], t] = runtimes[t][i][1]
        self.runtime_matrix = runtime_matrix

    def __check(self):
        if len(self.metrics) == 0:
            raise TypeError("No metrics. Use set_metrics(...) to set evaluation metrics before.")
        if self.instances is None:
            raise TypeError("No instances. Use set_instances(...) to set instances to be evaluated.")



import multiprocessing as mp

from sklearn.datasets import make_classification, make_multilabel_classification
from sklearn.model_selection import train_test_split

from pusion.util.processes import p_fit
from pusion.util.transformer import *


def generate_multiclass_ensemble_classification_outputs(classifiers, n_classes, n_samples, continuous_out=False,
                                                        parallelize=True):
    """
    Generate random multiclass, crisp and redundant classification outputs (assignments) for the given ensemble of
    classifiers.

    :param classifiers: Classifiers used to generate classification outputs.
            These need to implement `fit` and `predict` methods according to classifiers provided by `sklearn`.
    :param n_classes: `integer`. Number of classes, predictions are made for.
    :param n_samples: `integer`. Number of samples.
    :param parallelize: If `True`, all classifiers are trained in parallel. Otherwise they are trained in sequence.
    :param continuous_out: If `True`, class assignments in `y_ensemble_valid` and `y_ensemble_test` are given as
            probabilities. Default value is `False`.
    :return: `tuple` of:
            - `y_ensemble_valid`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a validation dataset.
            - `y_valid`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the validation.
            - `y_ensemble_test`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a test dataset.
            - `y_test`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the test.
    """
    x, y = make_classification(n_samples=n_samples,
                               n_classes=n_classes,
                               n_redundant=0,
                               n_informative=n_classes,
                               n_clusters_per_class=1)

    x_train, x_meta, y_train, y_meta = train_test_split(x, y, test_size=.5)
    x_valid, x_test, y_valid, y_test = train_test_split(x_meta, y_meta, test_size=.5)

    # Classifier training
    if parallelize:
        # Parallelized classifier training
        # Create a thread-safe queue for classifiers
        queue = mp.Manager().Queue()
        processes = []
        for i, clf in enumerate(classifiers):
            processes.append(mp.Process(target=p_fit, args=(i, clf, x_train, y_train, queue)))
        for p in processes:
            p.start()
        for p in processes:
            p.join()
        while not queue.empty():
            i, clf = queue.get()
            classifiers[i] = clf
    else:
        # Sequential classifier training
        for i, clf in enumerate(classifiers):
            print("Train classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            clf.fit(x_train, y_train)

    # Classifier validation to generate combiner training data
    y_ensemble_valid = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_valid.append(clf.predict_proba(x_valid))
    else:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_valid.append(clf.predict(x_valid))

    # Classifier test
    y_ensemble_test = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_test.append(clf.predict_proba(x_test))
    else:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_ensemble_test.append(clf.predict(x_test))

    # Transform to matrix representation
    if not continuous_out:
        y_ensemble_valid = transform_label_tensor_to_class_assignment_tensor(y_ensemble_valid, n_classes)
        y_ensemble_test = transform_label_tensor_to_class_assignment_tensor(y_ensemble_test, n_classes)
    y_valid = transform_label_vector_to_class_assignment_matrix(y_valid, n_classes)
    y_test = transform_label_vector_to_class_assignment_matrix(y_test, n_classes)

    # Transform to numpy tensors if possible
    y_ensemble_valid = tensorize(y_ensemble_valid)
    y_ensemble_test = tensorize(y_ensemble_test)

    return y_ensemble_valid, y_valid, y_ensemble_test, y_test


def generate_multiclass_cr_ensemble_classification_outputs(classifiers, n_classes, n_samples, coverage=None,
                                                           continuous_out=False, parallelize=True):
    """
    Generate random multiclass, crisp and complementary-redundant classification outputs (assignments) for the given
    ensemble of classifiers.

    :param classifiers: Classifiers used to generate classification outputs.
            These need to implement `fit` and `predict` methods according to classifiers provided by `sklearn`.
    :param n_classes: `integer`. Number of classes, predictions are made for.
    :param n_samples: `integer`. Number of samples.
    :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a classifier,
            which is identified by the positional index of the respective list.
            If unset, redundant classification outputs are retrieved.
    :param continuous_out: If `True`, class assignments in `y_ensemble_valid` and `y_ensemble_test` are given as
            probabilities. Default value is `False`.
    :param parallelize: If `True`, all classifiers are trained in parallel. Otherwise they are trained in sequence.
    :return: `tuple` of:
            - `y_ensemble_valid`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a validation dataset.
            - `y_valid`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the validation.
            - `y_ensemble_test`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a test dataset.
            - `y_test`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the test.
    """
    x, y = make_classification(n_samples=n_samples,
                               n_classes=n_classes,
                               n_redundant=0,
                               n_informative=n_classes,
                               n_clusters_per_class=1)

    x_train, x_meta, y_train, y_meta = train_test_split(x, y, test_size=.5)
    x_valid, x_test, y_valid, y_test = train_test_split(x_meta, y_meta, test_size=.5)

    # Transform to class assignments
    y_train = transform_label_vector_to_class_assignment_matrix(y_train, n_classes)
    y_valid = transform_label_vector_to_class_assignment_matrix(y_valid, n_classes)
    y_test = transform_label_vector_to_class_assignment_matrix(y_test, n_classes)

    # Classifier training
    if parallelize:
        # Parallelized classifier training
        # Create a thread-safe queue for classifiers
        queue = mp.Manager().Queue()
        processes = []
        for i, clf in enumerate(classifiers):
            y_train_ = y_train[:, coverage[i]]
            y_train_ = intercept_normal_class(y_train_, override=True)
            y_train_ = multiclass_assignments_to_labels(y_train_)
            processes.append(mp.Process(target=p_fit, args=(i, clf, x_train, y_train_, queue)))
        for p in processes:
            p.start()
        for p in processes:
            p.join()
        while not queue.empty():
            i, clf = queue.get()
            classifiers[i] = clf
    else:
        # Sequential classifier training
        for i, clf in enumerate(classifiers):
            print("Train classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_train_ = y_train[:, coverage[i]]
            y_train_ = intercept_normal_class(y_train_, override=True)
            y_train_ = multiclass_assignments_to_labels(y_train_)
            clf.fit(x_train, y_train_)

    # Classifier validation to generate combiner training data
    y_ensemble_valid = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict_proba(x_valid)
            y_ensemble_valid.append(y_pred)
    else:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict(x_valid)
            y_pred = transform_label_vector_to_class_assignment_matrix(y_pred, len(coverage[i]))
            y_ensemble_valid.append(y_pred)

    # Classifier test
    y_ensemble_test = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict_proba(x_test)
            y_ensemble_test.append(y_pred)
    else:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict(x_test)
            y_pred = transform_label_vector_to_class_assignment_matrix(y_pred, len(coverage[i]))
            y_ensemble_test.append(y_pred)

    # Transform to numpy tensors if possible
    y_ensemble_valid = tensorize(y_ensemble_valid)
    y_ensemble_test = tensorize(y_ensemble_test)

    return y_ensemble_valid, y_valid, y_ensemble_test, y_test


def generate_multilabel_ensemble_classification_outputs(classifiers, n_classes, n_samples,  continuous_out=False,
                                                        parallelize=True):
    """
    Generate random multilabel crisp classification outputs (assignments) for the given ensemble of classifiers with
    the normal class included at index `0`.

    :param classifiers: Classifiers used to generate classification outputs.
            These need to implement `fit` and `predict` methods according to classifiers provided by `sklearn`.
    :param n_classes: `integer`. Number of classes, predictions are made for with the normal class included.
    :param n_samples: `integer`. Number of samples.
    :param continuous_out: If `True`, class assignments in `y_ensemble_valid` and `y_ensemble_test` are given as
            probabilities. Default value is `False`.
    :param parallelize: If `True`, all classifiers are trained in parallel. Otherwise they are trained in sequence.
    :return: `tuple` of:
            - `y_ensemble_valid`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a validation dataset.
            - `y_valid`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the validation.
            - `y_ensemble_test`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a test dataset.
            - `y_test`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the test.
    """
    x, y = make_multilabel_classification(n_samples=n_samples,
                                          n_classes=n_classes-1,  # -1, due to normal class
                                          n_labels=2,
                                          allow_unlabeled=True)

    # Adapt y to address the normal class representation in case of an unlabeled output (prepend normal class)
    y = intercept_normal_class(y, override=False)

    x_train, x_meta, y_train, y_meta = train_test_split(x, y, test_size=.5)
    x_valid, x_test, y_valid, y_test = train_test_split(x_meta, y_meta, test_size=.5)

    # Classifier training
    if parallelize:
        # Parallelized classifier training
        # Create a thread-safe queue for classifiers
        queue = mp.Manager().Queue()
        processes = []
        for i, clf in enumerate(classifiers):
            processes.append(mp.Process(target=p_fit, args=(i, clf, x_train, y_train, queue)))
        for p in processes:
            p.start()
        for p in processes:
            p.join()
        while not queue.empty():
            i, clf = queue.get()
            classifiers[i] = clf
    else:
        # Sequential classifier training
        for i, clf in enumerate(classifiers):
            print("Train classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            clf.fit(x_train, y_train)

    # Classifier validation to generate combiner training data
    y_ensemble_valid = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict_proba(x_valid)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_valid.append(y_pred)
    else:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict(x_valid)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_valid.append(y_pred)

    # Classifier test
    y_ensemble_test = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict_proba(x_test)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_test.append(y_pred)
    else:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict(x_test)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_test.append(y_pred)

    # Transform to numpy tensors if possible
    y_ensemble_valid = tensorize(y_ensemble_valid)
    y_ensemble_test = tensorize(y_ensemble_test)

    return y_ensemble_valid, y_valid, y_ensemble_test, y_test


def generate_multilabel_cr_ensemble_classification_outputs(classifiers, n_classes, n_samples, coverage=None,
                                                           continuous_out=False, parallelize=True):
    """
    Generate random multilabel, crisp and complementary-redundant classification outputs (assignments) for the given
    ensemble of classifiers with the normal class included at index `0`.

    :param classifiers: Classifiers used to generate classification outputs.
            These need to implement `fit` and `predict` methods according to classifiers provided by `sklearn`.
    :param n_classes: `integer`. Number of classes, predictions are made for with the normal class included.
    :param n_samples: `integer`. Number of samples.
    :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a classifier,
            which is identified by the positional index of the respective list.
            If unset, redundant classification outputs are retrieved.
    :param continuous_out: If `True`, class assignments in `y_ensemble_valid` and `y_ensemble_test` are given as
            probabilities. Default value is `False`.
    :param parallelize: If `True`, all classifiers are trained in parallel. Otherwise they are trained in sequence.
    :return: `tuple` of:
            - `y_ensemble_valid`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a validation dataset.
            - `y_valid`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the validation.
            - `y_ensemble_test`: `numpy.array` of shape `(n_samples, n_classes)`. Ensemble decision output matrix for
            as a test dataset.
            - `y_test`: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments for the test.
    """
    x, y = make_multilabel_classification(n_samples=n_samples,
                                          n_classes=n_classes-1,  # -1, due to normal class
                                          n_labels=2,
                                          allow_unlabeled=True)

    # Adapt y to address the normal class representation in case of an unlabeled output (prepend normal class)
    y = intercept_normal_class(y, override=False)

    x_train, x_meta, y_train, y_meta = train_test_split(x, y, test_size=.5)
    x_valid, x_test, y_valid, y_test = train_test_split(x_meta, y_meta, test_size=.5)

    # Classifier training
    if parallelize:
        # Parallelized classifier training
        # Create a thread-safe queue for classifiers
        queue = mp.Manager().Queue()
        processes = []
        for i, clf in enumerate(classifiers):
            y_train_ = y_train[:, coverage[i]]
            y_train_ = intercept_normal_class(y_train_, override=True)
            processes.append(mp.Process(target=p_fit, args=(i, clf, x_train, y_train_, queue)))
        for p in processes:
            p.start()
        for p in processes:
            p.join()
        while not queue.empty():
            i, clf = queue.get()
            classifiers[i] = clf
    else:
        # Sequential classifier training
        for i, clf in enumerate(classifiers):
            print("Train classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_train_ = y_train[:, coverage[i]]
            y_train_ = intercept_normal_class(y_train_, override=True)
            clf.fit(x_train, y_train_)

    # Classifier validation to generate combiner training data
    y_ensemble_valid = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict_proba(x_valid)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_valid.append(y_pred)
    else:
        for i, clf in enumerate(classifiers):
            print("Validate classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict(x_valid)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_valid.append(y_pred)

    # Classifier test
    y_ensemble_test = []
    if continuous_out:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict_proba(x_test)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_test.append(y_pred)
    else:
        for i, clf in enumerate(classifiers):
            print("Test classifier: ", type(clf).__name__, " [" + str(i) + "] ...")
            y_pred = clf.predict(x_test)
            # y_pred = intercept_normal_class(y_pred, override=True)
            y_ensemble_test.append(y_pred)

    # Transform to numpy tensors if possible
    y_ensemble_valid = tensorize(y_ensemble_valid)
    y_ensemble_test = tensorize(y_ensemble_test)

    return y_ensemble_valid, y_valid, y_ensemble_test, y_test


def generate_multiclass_confusion_matrices(decision_tensor, true_assignments):
    """
    Generate multiclass confusion matrices out of the given decision tensor and true assignments.
    Continuous outputs are converted to multiclass assignments using the MAX rule.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of crisp decision outputs by different classifiers per sample.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which are considered true for calculating confusion matrices.
    :return: `numpy.array` of shape `(n_classifiers, n_samples, n_samples)`. Confusion matrices per classifier.
    """
    true_assignment_labels = np.argmax(true_assignments, axis=1)
    confusion_matrices = np.zeros((np.shape(decision_tensor)[0],
                                   np.shape(true_assignments)[1],
                                   np.shape(true_assignments)[1]), dtype=int)
    for i in range(len(decision_tensor)):
        decision_tensor_labels = np.argmax(decision_tensor[i], axis=1)
        confusion_matrices[i] = confusion_matrix(y_true=true_assignment_labels,
                                                 y_pred=decision_tensor_labels,
                                                 labels=np.arange(np.shape(true_assignments)[1]))
    return confusion_matrices


def generate_multilabel_cr_confusion_matrices(decision_outputs, true_assignments, coverage):
    """
    Generate multilabel confusion matrices for complementary-redundant multilabel classification outputs.

    :param decision_outputs: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
            `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
            due to the coverage.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which are considered true for calculating confusion matrices.
    :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a classifier,
            which is identified by the positional index of the respective list.
    :return: List of multilabel confusion matrices.
    """
    cr_confusion_matrices = []
    for i, do in enumerate(decision_outputs):
        ta = intercept_normal_class(true_assignments[:, coverage[i]], override=True)
        cms = multilabel_confusion_matrix(y_true=ta, y_pred=do, labels=np.arange(len(coverage[i])))
        cr_confusion_matrices.append(cms)
    return cr_confusion_matrices


def generate_classification_coverage(n_classifiers, n_classes, overlap, normal_class=True):
    """
    Generate random complementary redundant class indices for each classifier `0..(n_classifiers-1)`.
    The coverage is drawn from normal distribution for all classifiers.
    However, it is guaranteed that each classifier covers at least one class regardless of the distribution.

    :param n_classifiers: Number of classifiers representing the classifier `0..(n_classifiers-1)`.
    :param n_classes: Number of classes representing the class label `0..(n_classes-1)`.
    :param overlap: Indicator between `0` and `1` for overall classifier overlapping in terms of classes.
            If `0`, only complementary class indices are obtained.
            If `1`, the overlapping is fully redundant.
    :param normal_class: If `True`, a class for the normal state is included for all classifiers as class index `0`.
    :return: `list` of `list` elements. Each inner list contains classes as integers covered by a classifier,
            which is identified by the positional index of the respective list.
    """
    if normal_class:
        n_classes = n_classes - 1
    coverage_matrix = np.zeros((n_classifiers, n_classes), dtype=int)
    n_selected_classifier = int(np.interp(overlap, [0, 1], [np.ceil(n_classifiers/n_classes), n_classifiers]))
    while np.any(coverage_matrix.sum(axis=1) < 1):
        coverage_matrix = np.zeros((n_classifiers, n_classes), dtype=int)
        for i in range(n_classes):
            selected_classifier = np.random.choice(np.arange(n_classifiers), n_selected_classifier, replace=False)
            coverage_matrix[selected_classifier, i] = 1

    class_index_list = []
    if normal_class:
        for i in range(n_classifiers):
            class_index_list.append(np.array([0, *np.where(coverage_matrix[i])[0] + 1]))
    else:
        for i in range(n_classifiers):
            class_index_list.append(np.where(coverage_matrix[i])[0])
    return class_index_list


def shrink_to_coverage(decision_tensor, coverage):
    """
    Shrink the given decision tensor to decision outputs according to the given coverage.
    Assumption: the normal class is covered by each classifier at index `0`.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multilabel decision outputs by different classifiers per sample.
    :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a classifier,
            which is identified by the positional index of the respective list.
    :return: `list` of `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is
            classifier-specific due to the coverage.
    """
    decision_tensor = np.array(decision_tensor)
    decision_outputs = []
    for i in range(len(decision_tensor)):
        sdt = decision_tensor[i, :, coverage[i]].T
        # assign the normal class for samples with no assignment.
        sdt = intercept_normal_class(sdt, override=True)
        decision_outputs.append(sdt)
    return tensorize(decision_outputs)


def split_into_train_and_validation_data(decision_tensor, true_assignments, validation_size=0.5):
    """
    Split the decision outputs (tensor) from multiple classifiers as well as the true assignments randomly into train
    and validation datasets.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of decision outputs by different classifiers per sample.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of true class assignments.
    :param validation_size: Proportion between `0` and `1` for the size of the validation data set.
    :return: `tuple` of
            (1) `numpy.array` of shape `(n_classifiers, n_samples', n_classes)`,
            (2) `numpy.array` of shape `(n_classifiers, n_samples')`,
            (3) `numpy.array` of shape `(n_classifiers, n_samples'', n_classes)`,
            (4) `numpy.array` of shape `(n_classifiers, n_samples'')`, with `n_samples'` as the number of training
            samples and `n_samples''` as the number of validation samples.
    """
    n_validation_samples = int(len(true_assignments) * validation_size)
    all_indices = np.arange(len(true_assignments))
    validation_indices = np.random.choice(all_indices, n_validation_samples, replace=False)
    mask = np.ones(len(all_indices), bool)
    mask[validation_indices] = 0
    train_indices = all_indices[mask]
    true_assignments_train = true_assignments[train_indices]
    true_assignments_validation = true_assignments[validation_indices]
    decision_tensor_train = []
    decision_tensor_validation = []

    for decision_matrix in decision_tensor:
        decision_tensor_train.append(decision_matrix[train_indices])
        decision_tensor_validation.append(decision_matrix[validation_indices])

    return tensorize(decision_tensor_train), \
           tensorize(true_assignments_train), \
           tensorize(decision_tensor_validation), \
           tensorize(true_assignments_validation)



from pusion.core.combiner import TrainableCombiner
from pusion.util.transformer import *
from pusion.util.constants import *


class MaximumLikelihoodCombiner(TrainableCombiner):
    """
    The :class:`MaximumLikelihoodCombiner` (MLE) is a combiner that estimates the parameters :math:`\\mu` (sample means)
    and :math:`\\sigma` (sample variances) of the Gaussian probability density function for each class :math:`\\omega`.
    Multiple decision outputs for a sample are converted into a new feature space.

    The fusion is performed by evaluating the class conditional density

    .. math::
        p(x|\\omega) = \\frac{1}{\\sigma \\sqrt{2 \\pi}}
            exp\\left({-\\frac{1}{2}\\left(\\frac{x-\\mu}{\\sigma}\\right)^2}\\right).

    of a transformed sample :math:`x` for each available class :math:`\\omega`, respectively. The class with the highest
    likelihood is considered as winner and thus forms the decision fusion.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'MLE'

    def __init__(self):
        TrainableCombiner.__init__(self)
        self.unique_assignments = None
        self.mu = []
        self.sigma = []

    def train(self, decision_tensor, true_assignments):
        """
        Train the Maximum Likelihood combiner model by calculating the parameters of gaussian normal distribution
        (i.e. means and variances) from the given decision outputs and true class assignments.
        Both continuous and crisp classification outputs are supported. This procedure transforms decision outputs
        into a new feature space.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        # transfer decisions into a new feature space
        featured_decisions = decision_profiles.reshape((decision_profiles.shape[0], -1))

        # extract all occurring classifications
        self.unique_assignments, unique_inv_indices = np.unique(true_assignments, axis=0, return_inverse=True)

        # calculate the parameters for the multivariate normal distribution
        for i in range(len(self.unique_assignments)):
            xc = featured_decisions[np.where(unique_inv_indices == i)]  # X_train(class)
            mu = np.mean(xc, axis=0)
            sigma = np.std(xc, axis=0)
            sigma[sigma == 0] = 0.00001  # Add a small perturbation in order to enable class conditional density

            self.mu.append(mu)
            self.sigma.append(sigma)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by the Maximum Likelihood method. This procedure involves evaluating the class
        conditional density as described above. Both continuous and crisp classification outputs are supported.
        Combining requires a trained :class:`MaximumLikelihoodCombiner`.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by MLE. Axis 0 represents samples and axis 1 the class assignments which are aligned
                with axis 2 in ``decision_tensor`` input tensor.
        """
        fused_decisions = np.zeros_like(decision_tensor[0])
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        # transfer decisions into a new feature space
        featured_decisions = decision_profiles.reshape((decision_profiles.shape[0], -1))

        for i in range(len(featured_decisions)):
            x = featured_decisions[i]
            likelihoods = np.ones(len(self.unique_assignments))
            for j in range(len(self.unique_assignments)):
                for k in range(len(x)):
                    # calculate class conditional density for each dimension
                    exp = (x[k] - self.mu[j][k])/self.sigma[j][k]
                    likelihoods[j] = likelihoods[j] * 1/(np.sqrt(2*np.pi) * self.sigma[j][k]) * np.e**(-.5*(exp**2))
            fused_decisions[i] = self.unique_assignments[np.argmax(likelihoods)]
        return fused_decisions


class CRMaximumLikelihoodCombiner(MaximumLikelihoodCombiner):
    """
    The :class:`CRMaximumLikelihoodCombiner` is a modification of :class:`MaximumLikelihoodCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as a constant, respectively. To use methods :meth:`train` and
    :meth:`combine` a coverage needs to be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        # (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT), # performance issues
        # (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT), # performance issues
    ]

    def __init__(self):
        super().__init__()

    def train(self, decision_outputs, true_assignments):
        """
        Train the Maximum Likelihood combiner model by calculating the parameters of gaussian normal distribution
        (i.e. means and variances) from the given decision outputs and true class assignments.
        Both continuous and crisp classification outputs are supported. This procedure transforms decision outputs
        into a new feature space.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains either crisp or continuous
                decision outputs per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        super().train(t_decision_outputs, true_assignments)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by the Maximum Likelihood method. This procedure involves evaluating the class
        conditional density as described above. Both continuous and crisp classification outputs are supported.
        Combining requires a trained :class:`MaximumLikelihoodCombiner`.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by MLE. Axis 0 represents samples and axis 1 the class assignments which are aligned
                with axis 2 in ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.negative(np.ones((n_classifiers, n_decisions, n_classes)))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



from sklearn.neural_network import MLPClassifier

from pusion.core.combiner import TrainableCombiner
from pusion.util.transformer import *
from pusion.util.constants import *


class NeuralNetworkCombiner(TrainableCombiner):
    """
    The :class:`NeuralNetworkCombiner` (NN) is a learning and classifier-based combiner that converts multiple decision
    outputs into new features, which in turn are used to train this combiner. The NN includes three hidden layers and a
    dynamic number of neurons per layer, which is given by (`n_classifiers * n_classes`).
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'NN'

    N_LAYERS = 3

    def __init__(self):
        TrainableCombiner.__init__(self)
        self.classifier = None

    def train(self, decision_tensor, true_assignments):
        """
        Train the NN combiner by fitting the Neural Network model with given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported.
        This procedure transforms decision outputs into a new feature space.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        hidden_layer_sizes = (true_assignments.shape[1]*len(decision_tensor),) * self.N_LAYERS
        self.classifier = MLPClassifier(max_iter=5000, hidden_layer_sizes=hidden_layer_sizes)

        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        # transfer decisions into a new feature space
        featured_decisions = decision_profiles.reshape((decision_profiles.shape[0], -1))

        self.classifier.fit(featured_decisions, true_assignments)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by the trained Neural Network model. Both continuous and crisp classification outputs
        are supported. Combining requires a trained :class:`NeuralNetworkCombiner`. This procedure transforms decision
        outputs into a new feature space.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by NN. Axis 0 represents samples and axis 1 the class assignments which are aligned
                with axis 2 in ``decision_tensor`` input tensor.
        """
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        # transfer decisions into a new feature space
        featured_decisions = decision_profiles.reshape((decision_profiles.shape[0], -1))

        # return self.classifier.predict_proba(featured_decisions)  # continuous output
        return self.classifier.predict(featured_decisions)


class CRNeuralNetworkCombiner(NeuralNetworkCombiner):
    """
    The :class:`CRNeuralNetworkCombiner` is a modification of :class:`NeuralNetworkCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as a constant, respectively. To use methods :meth:`train` and
    :meth:`combine` a coverage needs to be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT)
    ]

    def __init__(self):
        super().__init__()

    def train(self, decision_outputs, true_assignments):
        """
        Train the NN combiner by fitting the Neural Network model with given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported.
        This procedure transforms decision outputs into a new feature space.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains either crisp or continuous
                decision outputs per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        super().train(t_decision_outputs, true_assignments)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by the trained Neural Network model. Both continuous and crisp classification outputs
        are supported. Combining requires a trained :class:`NeuralNetworkCombiner`. This procedure transforms decision
        outputs into a new feature space.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by NN. Axis 0 represents samples and axis 1 the class assignments which are aligned
                with axis 2 in ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.negative(np.ones((n_classifiers, n_decisions, n_classes)))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



from pusion.core.combiner import TrainableCombiner, EvidenceBasedCombiner
from pusion.util.generator import *
from pusion.util.constants import *


class NaiveBayesCombiner(EvidenceBasedCombiner, TrainableCombiner):
    """
    The :class:`NaiveBayesCombiner` (NB) is a fusion method based on the Bayes theorem which is applied according to
    Kuncheva :footcite:`kuncheva2014combining` and Titterington et al. :footcite:`titterington1981comparison`.
    NB uses the confusion matrix as an evidence to calculate the a-priori probability and the bayesian belief value,
    which in turn the decision fusion bases on. NB requires outputs from uncorrelated classifiers in the ensemble.

    .. footbibliography::
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT)
    ]

    SHORT_NAME = 'NB'

    def __init__(self):
        TrainableCombiner.__init__(self)
        self.confusion_matrices = None
        # Number of samples per class
        self.n_samples_per_class = None

    def set_evidence(self, evidence):
        """
        Set the evidence given by confusion matrices calculated according to Kuncheva :footcite:`kuncheva2014combining`
        for each ensemble classifier.

        .. footbibliography::

        :param evidence: `numpy.array` of shape `(n_classifiers, n_classes, n_classes)`.
                Confusion matrices for each of `n` classifiers.
        """
        self.confusion_matrices = evidence
        self.n_samples_per_class = np.sum(evidence[0], axis=1)

    def train(self, decision_tensor, true_assignments):
        """
        Train the Naive Bayes combiner model by precalculating confusion matrices from given decision outputs and
        true class assignments. Continuous decision outputs are converted into crisp multiclass assignments using
        the MAX rule.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        self.confusion_matrices = generate_multiclass_confusion_matrices(decision_tensor, true_assignments)
        # Number of samples of certain class (multiclass case)
        self.n_samples_per_class = np.sum(true_assignments, axis=0)
        # self.sum_of_samples_per_class = np.sum(confusion_matrices[0], axis=1)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by using the Naive Bayes method.
        Continuous decision outputs are converted to crisp multiclass predictions using the MAX rule.
        Combining requires a trained :class:`NaiveBayesCombiner` or evidence set with ``set_evidence``.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which represents fused
                decisions obtained by the maximum class support. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        if self.confusion_matrices is None or self.n_samples_per_class is None:
            raise RuntimeError("Untrained model.")
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        fused_decisions = np.zeros_like(decision_tensor[0])
        for i in range(len(decision_profiles)):
            # transform to a multiclass decision profile
            dp = multiclass_predictions_to_decisions(decision_profiles[i])
            n_classes = dp.shape[1]
            mu = np.zeros(n_classes)
            for j in range(n_classes):
                n_j = self.n_samples_per_class[j]
                mu[j] = n_j / len(decision_profiles)
                for k in range(dp.shape[0]):
                    mu[j] = mu[j] * ((self.confusion_matrices[k, j, np.argmax(dp[k])] + 1/n_classes) / (n_j + 1))
            fused_decisions[i, np.argmax(mu)] = 1
        return fused_decisions


class CRNaiveBayesCombiner(NaiveBayesCombiner):
    """
    The :class:`CRNaiveBayesCombiner` is a modification of :class:`NaiveBayesCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as `0`, respectively. To call :meth:`combine` a coverage needs to be set
    first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT)
    ]

    def __init__(self):
        super().__init__()

    def train(self, decision_outputs, true_assignments):
        """
        Train the Naive Bayes combiner model by precalculating confusion matrices from given decision outputs and
        true class assignments. Continuous decision outputs are converted into crisp multiclass assignments using
        the MAX rule.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains either crisp or continuous
                decision outputs per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        super().train(t_decision_outputs, true_assignments)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by using the Naive Bayes method.
        Continuous decision outputs are converted to crisp multiclass predictions using the MAX rule.
        Combining requires a trained :class:`NaiveBayesCombiner` or evidence set with ``set_evidence``.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        # use zeros to generate proper confusion matrices in the training phase
        t_decision_outputs = np.zeros((n_classifiers, n_decisions, n_classes))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs
